import os
import numpy as np
import matplotlib
from hylite.correct.detrend import get_hull_corrected
from gfit import initialise, gfit,evaluate
from hylite import HyCollection, HyCloud, HyImage, HyData

import matplotlib.pyplot as plt
import matplotlib as mpl
from tqdm import tqdm

class MWL(HyCollection):
    """
    A convenient class for manipulating and storing minimum wavelength mapping results.
    """

    def __init__(self, name, root, header=None):
        super().__init__(name, root, header)
        self.ext = '.mwl'

    def bind(self, model, nfeatures, x, X, sym=False):
        """
        Band a new MWL mapping results. Essentially treat this as the constructor.

        Args:
            model: the underlying HyData instance containing feature parameters.
            nfeatures: the number of features stored in the underlying model.
            x: Wavelengths that the model should be evaluated at.
            X: the underlying data that the model was fitted to. Useful for plotting / debugging. Default is None.
            sym: True if symmetric features are stored (3-parameters). Default is False (4-parameters).
        """

        # store attributes
        self.model = model  # store underlying feature model
        self.x = x # model domain
        self.X = X # fitted data
        self.n = nfeatures  # store number of features
        self.sym = sym # store symmetry
        if sym: # set stride
            self.stride = 3
        else:
            self.stride = 4

    def getAttributes(self):
        """
        Return a list of available attributes in this HyScene. We must override the HyCollection implementation to remove
        functions associated with HyScene.
        """
        return list(set(dir(self)) - set(dir(HyCollection)) - set(dir(MWL)) - set(['header', 'root', 'name']))

    def __getitem__(self, n):
        """
        Slice this MWL object to return specific features or feature parameters.

        Options are:
         self[n]: get the n'th feature (as a HyData instance). See self.sortByDepth(..) and self.sortByPos(..) to
                   change feature order.
         self[n,b]: return a numpy array containing a specific property of the n'th feature. b can be an index 0-3
                     for symmetric and 0-4 for asymmetric features, or string ('depth', 'pos', 'width', 'width2').
        """

        # return MWL as HyData instance
        if isinstance(n, int):
            assert n < self.n, "Error - MWL has only %d features (not %d)" % (self.n, n + 1)
            out = self.model.copy(data=False)
            out.data = self.model.data[...,
                       n * self.stride:(n + 1) * self.stride]  # return bands associated with the features
            return out
        else:  # return slice of MWL data as numpy array.
            assert len(n) == 2, "Error - %s is an invalid key." % (n)
            #  parse parameter descriptor
            if isinstance(n[1], str):
                if 'depth' in n[1]:
                    b = 0
                if 'pos' in n[1]:
                    b = 1
                elif 'width2' in n[1]:
                    b = 3
                    assert not self.sym, "Error - symmetric mwl features have no width2."
                elif 'width' in n[1]:
                    b = 2
            elif isinstance(n[1], slice):
                b = n[1]  # no change needed
            else:
                assert isinstance(n[1], int), "Error - key should be integer"
                b = n[1]
                if self.sym:
                    assert b < 2, "Error - %d is an invalid (symmetric) mwl band index. Should be 0 - 2."
                else:
                    assert b < 3, "Error - %d is an invalid mwl band index. Should be 0 - 3."

            if isinstance(n[0], slice):  # we want to slice only depth or only
                assert not isinstance(b, slice), "Error - invalid slice key %s" % (n)
                return self.model.data[..., b::self.stride]
            else:
                assert n[0] < self.n, "Error - MWL has only %d features (not %d)" % (self.n, n[0] + 1)
                return self.model.data[..., n[0] * self.stride + b]

    def getFeature(self, n):
        """
        Return a HyData instance containing bands associated with th nth minimum wavelength feature.
        """
        return self[n]

    def sortByDepth(self):
        """
        Sort features such that they are stored/returned from deepest to shallowest.
        """
        # sort by depth
        depth = self.model.X()[:, ::self.stride]
        idx = np.argsort(-depth, axis=-1)

        # expand index
        idx_full = np.zeros((depth.shape[0], self.model.band_count()), dtype=int)
        for c in range(self.stride):
            for n in range(idx.shape[-1]):
                idx_full[:, n * self.stride + c] = idx[:, n] * self.stride + c

        # take values to rearranged form
        out = np.take_along_axis(self.model.X(), idx_full, axis=-1).reshape(self.model.data.shape)
        self.model.set_raveled(out)

    def sortByPos(self):
        """
        Sort features such that they are stored/returned from lowest to highest wavelength.
        """
        # sort by depth
        pos = self.model.X()[:, 1::self.stride]
        idx = np.argsort(pos, axis=-1)

        # expand index
        idx_full = np.zeros((pos.shape[0], self.model.band_count()), dtype=int)
        for c in range(self.stride):
            for n in range(idx.shape[-1]):
                idx_full[:, n * self.stride + c] = idx[:, n] * self.stride + c

        # take values to rearranged form
        out = np.take_along_axis(self.model.X(), idx_full, axis=-1).reshape(self.model.data.shape)
        self.model.set_raveled(out)

    def deepest(self, wmin=0, wmax=-1):
        """
        Returns a HyData instance containing the deepest feature within the specified range.

        Args:
            wmin: the lower bound of the wavelength range. Default is 0 (accept all positions)
            wmax: the upper bound of the wavelength range. Default is -1 (accept all positions)
        Returns:
            a MWL instance containing the deepest features within the range, or nans if no feature exists.
        """

        if wmin==0:
            wmin = np.nanmin(self[:,'pos'])
        if wmax==-1:
            wmax = np.nanmax(self[:,'pos'])

        # get valid positions
        valid_pos = (self[:, 'pos'] > wmin) & (np.array(self[:, 'pos']) < wmax)

        # get depths and filter to only include valid positions
        depth = self[:, 'depth'].copy()
        depth[np.logical_not(valid_pos)] = 0
        depth = depth.reshape((-1, self.n))

        # find deepest feature and expand index
        idx = np.argmax(depth, axis=-1)
        idx_full = np.zeros((depth.shape[0], self.stride), dtype=int)
        for c in range(self.stride):
            idx_full[:, c] = idx * self.stride + c

        # return deepest feature
        out = self.model.copy(data=False)
        out.data = np.take_along_axis(self.model.X(), idx_full, axis=-1).reshape(
            self.model.data.shape[:-1] + (-1,)).copy()
        out.data[out.data[..., 0] == 0, :] = np.nan  # remove 0 depths
        out.data[np.logical_not(valid_pos.any(axis=-1)), :] = np.nan  # remove invalid positions
        return out

    def closest(self, position, valid_range=None, depth_cutoff=0.05):
        """
        Returns a HyData instance containing the closest feature within the specified range.

        Args:
            position: the 'ideal' feature position to compare with (e.g. 2200.0 for AlOH)
            valid_range: A tuple defining the minimum and maximum acceptible wavelengths, or None (default). Values outside
                         of the valid range will be set to nan.
            depth_cutoff: Features with depths below this value will be discared (and set to nan).
        Returns:
            a single HyData instance containing the closest minima.
        """
        # get valid positions
        valid_pos = np.isfinite(self[:, 'pos'])
        if valid_range is not None:
            valid_pos = (self[:, 'pos'] > valid_range[0]) & (np.array(self[:, 'pos']) < valid_range[1])

        # get deviations and filter to only include valid positions
        dp = self[:, 'pos'].copy() - position
        dp[np.logical_not(valid_pos)] = np.inf
        dp = dp.reshape((-1, self.n))

        # find closest feature and expand index
        idx = np.argmin(dp, axis=-1)
        idx_full = np.zeros((dp.shape[0], self.stride), dtype=int)
        for c in range(self.stride):
            idx_full[:, c] = idx * self.stride + c

        # return closest feature
        out = self.model.copy(data=False)
        out.data = np.take_along_axis(self.model.X(), idx_full, axis=-1).reshape(
            self.model.data.shape[:-1] + (-1,)).copy()
        out.data[out.data[..., 0] == 0, :] = np.nan  # remove 0 depths
        out.data[np.logical_not(valid_pos.any(axis=-1)), :] = np.nan  # remove invalid positions
        return out

    def feature_between(self, wmin, wmax, depth_cutoff=0.05):
        """
        Return True if the entries in each pixel/point include a feature within the specified wavelength range,
        and False otherwise. Useful for e.g. decision tree classifications.

        Args:
            wmin: the lower bound of the wavelength range.
            wmax: the upper bound of the wavelength range.
            depth_cutoff: the minimum depth required for a feature to count as existing.

        Returns:
            a numpy array  populated with False (no feature) or True (feature).
        """
        valid_pos = (self[:, 'pos'] > wmin) & (np.array(self[:, 'pos']) < wmax)
        valid_depth = (self[:, 'depth'] > depth_cutoff)
        return (valid_pos & valid_depth).any(axis=-1)

    def evaluate(self):
        """
        Evaluate this model and return the result as a HyData instance.

        Returns:
            A HyData instance containing the estimated spectra based on the fitted features.
        """
        out = self.model.copy(data=False)
        out.data = 1. - evaluate(self.x, self.model.data, sym=self.sym)
        out.set_wavelengths(self.x)
        return out

    def residual(self, sum=False):
        """
        Evaluate and return the residuals to the fitted minimum wavelength model.

        Returns:
            A HyData instance containing the residuals in band 0.
        """
        out = self.model.copy(data=False)
        out.data = np.sum(np.abs(self.X.data - self.evaluate().data), axis=-1)[..., None]
        return out

    def classify(self, n, nf=2, step=1):
        """
        Identify clusters in feature position space to classify this MWL map.
        This uses the hierarchichal method scipy.cluster.hierarchy.fclusterdata.

        See this publication for more details: https://doi.org/10.3390/min11020136

        Args:
            n: the number of classes to use.
            nf: the number of feature positions to use. Default is 2. Must not exceed the number of features fitted.
            step: the step to subsample points (used to avoid really slow plotting). Computes on all points by default.
        Returns:
            A tuple containing:

            - labels =  a HyData instance (or numpy array if step > 1) containing integer class labels in band 0.
            - centroids = a list containing the index of each class centroid (in the dataset).
         """

        assert nf <= self.n, "Error - MWL map has only %d features (<%d)." % (self.n, nf)
        import scipy.cluster.hierarchy as shc
        X = self[:, 'pos'][..., 0:nf]
        idx = np.array(np.meshgrid(*[np.arange(i) for i in self.X.data.shape[:-1]])).T

        # generate mask that subsamples and removes nans
        mask = np.full(X.shape[:-1], False)
        mask.ravel()[::step] = True  # subsample
        mask = mask & np.isfinite(X).all(axis=-1)  # remove nans
        X = X[mask, :]
        idx = idx[mask, :]

        # do clustering and get class centroids (in MWL space)
        L = shc.fclusterdata(X, n, criterion='maxclust', method='ward')
        C = np.array([np.median(X[L == i, :], axis=0) for i in range(1, n + 1)])

        # get index of centroids
        Cn = np.array([idx[np.argmin(np.linalg.norm(X - c, axis=1)), :] for c in C])  # center indices

        # return results
        if step == 1:
            out = self.model.copy(data=False)
            out.data = np.full(self[:, 'pos'].shape[:-1] + (1,), np.nan)
            out.data[mask, 0] = L
        else:
            out = np.full( mask.shape, np.nan )
            out[mask] = L

        return out, Cn


    ####################################
    ## Plotting functions
    ####################################
    def plot_features(self, ax=None, **kwds):
        """
        Plot all features in this minimum wavelength model on a scatter plot.

        Args:
            ax: a different axes to plot this figure on. Default is None (creates a new axis).
            **kwds: Keywords can include:

                 - step = the step to subsample points (to avoid really slow plotting). Defaults to a step that gives 1000 points.
                 - n = the number of classes to use for classification (see self.classify()),
                     or a list of class ids as returned by classify.
                 - cmap = the colour map to use (string). Default is 'tab10'.
                 - point_size = the size of the points to plot.
                 - point_alpha = the transparency of the points to plot.
                 - legend = True if a legend should be plotted. Default is True.
        Returns:
            fig,ax = the figure that was plotted.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))
        fig = ax.get_figure()

        ax.set_title('Spectral feature summary')

        # get data
        symbols = ['s', 'v', 'o', '.']
        names = ['Primary', 'Secondary', 'Tertiary', 'Other']
        p = self[:, 'pos']
        d = self[:, 'depth']

        # compute step
        step = int(kwds.get("step", np.prod(p.shape[:-1]) / 5000.))
        step = max(step, 1)  # step cannot be < 1

        # compute colours
        n = kwds.get('n', 5)
        if isinstance(n, int):
            L, _ = self.classify(n, step=step)  # self.classify(n, step=step)
            if step == 1:
                n = L.X()
            else:
                n = L

        c = mpl.cm.get_cmap(kwds.get('cmap', 'tab10'))(n.ravel() / np.nanmax(n))

        for i, f in enumerate(range(self.n)):
            if i >= len(symbols):
                i = -1
            ax.scatter(p[..., f].ravel()[0], d[..., f].ravel()[0], c='k', marker=symbols[i],
                       label='%s' % names[i], zorder=-1)  # plot single point for legend
            #print(c.shape, n.shape, n.ravel().shape, '   ', p[...,f].ravel().shape, p[...,f].ravel()[::step].shape)
            ax.scatter(p[..., f].ravel()[::step], d[..., f].ravel()[::step], c=c[::step],
                       marker=symbols[i],
                       s=kwds.get("point_size", 20), alpha=kwds.get("point_alpha", 0.5), lw=0)

            ax.set_title("Spectral feature summary")
            ax.set_xlabel("Feature position (nm)")
            ax.set_ylabel("Hull-corrected depth")
        if kwds.get('legend', True):
            ax.legend()

        return fig, ax

    def biplot(self, f1=0, f2=1, ax=None, **kwds):
        """
        Plot all features in this minimum wavelength model on a scatter plot. Note that calling
        this function will sort the features in this MWL instance by depth.

        Args:
            f1: the index of the first feature to plot (sorted by depth). Default is 0 (deepest feature).
            f2: the index of the first feature to plot (sorted by depth). Default is 1.
            ax: a different axes to plot this figure on. Default is None (creates a new axis).
            **kwds: Keywords can include:

                - step = the step to subsample points (to avoid really slow plotting). Defaults to a step that gives 1000 points.
                - n = the number of classes to use for classification (see self.classify()), or a list of class ids as returned by classify.
                - cmap = the colour map to use (string). Default is 'tab10'.
                - point_size = the size of the points to plot.
                - point_alpha = the transparency of the points to plot.
                - legend = True if a legend should be plotted. Default is True.
        Returns:
            fig,ax = the figure that was plotted.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))
        fig = ax.get_figure()

        ax.set_title('Spectral feature summary')

        # get feature information
        self.sortByDepth()
        p = self[:, 'pos']

        # compute step
        step = int(kwds.get("step", np.prod(p.shape[:-1]) / 5000.))
        step = max(step, 1)  # step cannot be < 1

        # compute colours
        n = kwds.get('n', 5)
        if isinstance(n, int):
            L, _ = self.classify(n, step=step)  # self.classify(n, step=step)
            if step == 1:
                n = L.X()
            else:
                n = L

        c = mpl.cm.get_cmap(kwds.get('cmap', 'tab10'))(n.ravel() / np.nanmax(n))

        # draw biplot
        ax.scatter(p[..., 0].ravel()[::step], p[..., 1].ravel()[::step], c=c[::step], lw=0, s=kwds.get("point_size", 20),
                   alpha=kwds.get("point_alpha", 0.5))  # plot

        ax.set_title("Deepest feature biplot")
        ax.set_xlabel("Deepest feature position (nm)")
        ax.set_ylabel("Secondary feature position (nm)")

        return fig, ax

    def quick_plot(self, **kwds):
        """
        Plot an overview visualisation that is useful for QAQC of MWL mapping results.

        Args:
            **kwds: Keywords can include:

                 - image = the image preview to plot. Can be a HyData instance, or 'resid' (default) to plots the average residuals,
                           or 'class' to a classification [slow!].
                 - bands = the bands of image to plot (if image is provided).
                 - vmin = the vmin value for plotting 'image' (if image is provided).
                 - vmax = the vmax value for plotting 'image' (if image is provided).
                 - residual_clip = the (mn,mx) percentile colour stretch to apply to the residuals.
                 - n = the number of clusters do create during absorbtion-feature clustering, or a list with precalculated
                            cluster labels.
                 - nf = the number of features to use for this clustering. Default is 2.
                 - offset = the vertical offset between spectra in the spectral summaries. Default is 0.25
                 - cmap = the colour map to use (string). Default is 'tab10'.
                 - point_size = the size of the points to plot.
                 - point_alpha = the transparency of the points to plot.
                 - legend = True if a legend should be plotted. Default is True.
                 - cam = a camera instance if the underlying dataset is a pointcloud. Default is 'ortho'.
                 - s = specify point size if the underlying dataset is a pointcloud. Default is 1.
        Returns:
            fig,ax = the plot that was created.
        """

        # todo; write another function for single-feature maps?
        assert self.n >= 2, "Error - quick_plot(...) can only be used for MWL instances with > 2 features."

        fig = plt.figure(constrained_layout=True, figsize=kwds.get('figsize', (15, 10)))
        gs = fig.add_gridspec(3, 2)
        ax1 = fig.add_subplot(gs[0, :])
        ax2a = fig.add_subplot(gs[1, 0])
        ax2b = fig.add_subplot(gs[1, 1])
        ax3a = fig.add_subplot(gs[2, 0])
        ax3b = fig.add_subplot(gs[2, 1])

        # compute residuals
        E = self.residual()
        mn, mx = np.nanpercentile(E.data, kwds.get("residual_clip", (2, 98)))

        # compute step
        step = int(kwds.get("step", np.prod(self.model.data.shape[:-1]) / 5000.))
        step = max(step, 1)  # step cannot be < 1
        if isinstance(kwds.get('image','resid'), str) and 'class' in kwds.get('image', 'resid'):
            step = 1 # step must be one if we want to plot the classification
        kwds["step"] = step  # store for later functions

        # do clustering
        n = kwds.get('n', 5)
        if isinstance(n, int):
            if step == 1: # compute full clustering... (slow)
                if E.X().shape[0] > 5000: # warn about speed
                    print("Computing clusters... this can be slow.")
                L, Cn = self.classify(n, kwds.get('nf', 2))
                kwds['n'] = L.X().ravel()
            else:
                L, Cn = self.classify(n, kwds.get('nf', 2), step=step)
                kwds['n'] = L.ravel()

        # get colormap
        cmap = mpl.cm.get_cmap(kwds.get('cmap', 'tab10'))

        # plot overview image
        ax1.set_xticks([])
        ax1.set_yticks([])
        image = kwds.get('image', 'resid')
        if isinstance(image, HyData): # image has been provided
            bands = kwds.get('bands', (0,1,2))
            vmin = kwds.get('vmin', 0)
            vmax = kwds.get('vmax', 99)
            if isinstance(image, HyImage):
                image.quick_plot(bands, ax=ax1, vmin=vmin, vmax=vmax)  # plot
            elif isinstance(image, HyCloud):
                image.quick_plot(bands, cam=kwds.get('cam', 'ortho'), s=kwds.get('s', 1),
                             ax=ax1, vmin=vmin, vmax=vmax)
        elif 'class' in image:  # plot classification
            assert step == 1, "Error - to plot classification, step must = 1."
            if isinstance(self.model, HyImage):
                L.quick_plot(0, cmap=cmap, ax=ax1, vmin=0, vmax=np.nanmax(kwds['n']))  # plot
            elif isinstance(self.model, HyCloud):
                L.quick_plot(0, cam=kwds.get('cam', 'ortho'), s=kwds.get('s', 1), cmap=cmap,
                             ax=ax1, vmin=0, vmax=np.nanmax(kwds['n']))
            ax1.set_title("Classification")
        elif 'resid' in image:  # plot residuals
            if isinstance(self.model, HyImage):
                E.quick_plot(0, cmap='gray', vmin=mn, vmax=mx, ax=ax1)
            elif isinstance(self.model, HyCloud):
                E.quick_plot(0, cam=kwds.get('cam', 'ortho'), s=kwds.get('s', 1),
                             cmap='gray', vmin=mn, vmax=mx, ax=ax1)
            ax1.set_title("Residuals")
        else:
            assert False, 'Error = %s is an unknown image type.' % image

        # position / depth plots
        self.plot_features(ax=ax2a, **kwds)

        # biplot
        self.biplot(ax=ax2b, **kwds)

        # biggest residual spectra
        ax3a.set_title('Spectral fitting quality')
        offs = 0
        mm = self.evaluate()
        for p, c, n in zip((50, 75, 95, 99), ['lightgreen', 'skyblue', 'orange', 'red'],
                           ['typical', 'dodgy', 'bad', 'worst']):
            t = np.nanpercentile(E.X(), p)
            idx = np.argmin(np.abs(np.nan_to_num(E.X()[:, 0] - t, nan=np.inf)))
            ax3a.plot(self.x, self.X.X()[idx, :] + offs, color=c, lw=2, label=n)
            ax3a.plot(self.x, mm.X()[idx, :] + offs, color=c, ls=':')
            offs += kwds.get('offset', 0.25)
        ax3a.legend()
        ax3a.set_xlabel("Wavelength (nm)")
        ax3a.set_ylabel("Hull-corrected reflectance")

        # plot cluster centroids
        offs = 0
        for i, idx in enumerate(Cn):
            c = cmap((i + 1) / np.nanmax(kwds['n']))

            # plot spectra on main image?
            if isinstance(self.model, HyImage):
                ax1.scatter(idx[0], idx[1], color=c, marker='o', edgecolors='k', lw=1)

            # stack spectra and plot
            y = self.X.data[ tuple(idx) ]
            ax3b.plot(self.x, y+offs, color=c, lw=1.7, label=n, alpha=0.8)
            y = mm.data[ tuple(idx)]
            ax3b.plot(self.x, y + offs, color=c, ls='--', alpha=0.8)

            offs += kwds.get('offset', 0.25)

        ax3b.set_title("Cluster centroids")
        ax3b.set_xlabel("Wavelength (nm)")
        ax3b.set_ylabel("Hull-corrected reflectance")

        return fig, [ax1, ax2a, ax2b, ax3a, ax3b]

    def plot_spectra(self, indices, ax=None, **kwds):
        """
        Plot the fitted features (and underlying data) for the specified indices.

        Args:
            indices: a list containing the indices to plot.
            ax: a matplotlib axes to plot to, or None (default) to create a new one.
            **kwds: Keywords can include:

                 - offset = the vertical offset between successive spectra. Default is 0.25.
                 - leg = True if a legend should be plotted. Default is True.
        """

        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 5))
        fig = ax.get_figure()
        ax.set_title('Fitted spectra')
        offs = 0

        for i, idx in enumerate(indices):

            # evaluate model and plot it
            if np.issubdtype(type(idx), int):
                y = 1 - evaluate(self.x, self.model[idx, :], self.sym)
            else:
                y = 1 - evaluate(self.x, self.model[idx[0], idx[1], :], self.sym)
            ax.plot(self.x, y + offs, color=plt.cm.tab10(i), ls='--', lw=2, label='Spectra %s' % (str(idx)))

            # plot data (if it is defined)
            try:
                if np.issubdtype(type(idx), int):
                    y = self.X[idx, :]
                else:
                    y = self.X[idx[0], idx[1], :]
                ax.plot(self.x, y + offs, color=plt.cm.tab10(i), lw=2, ls='-')
            except:
                continue

            offs += kwds.get('offset', 0.25)

        if kwds.get('leg', True):
            ax.legend()
        ax.set_xlabel("Wavelength (nm)")
        return fig, ax

def _refine(x, X, x0, n=3, vb=False):
    """
    Use a best-fit quadratic parabola to refine estimates of absorption feature position (as returned by init(...).
    N.B. this is much faster than gaussian fitting!

    :param x: = a (n,) array containing the x-values of the spectra / signals.
    :param X: = a (m,n) array containing corresponding y-values for m different spectra / signals.
    :param x0: = an array of estimated gaussian functions based on local peak detection as returned by init(...).
    :param n: = the number of bands on either side of the minima to use for fitting. Default is 5. Must be > 2.
    :param vb: = True if a progress bar should be created. Default is False.
    """

    # init output
    x1 = x0.copy()

    # loop through n maxima
    loop = range(1, x0.shape[-1], 4)
    if vb:
        loop = tqdm(loop, desc="Refining features", leave=False )
    for j in loop:
        # get fitting interval
        b = np.argmin(np.abs(x0[:, j, None] - x[None, :]), axis=1)
        b0 = np.clip(b - n, 0, x.shape[0] - 1)
        b1 = np.clip(b + n, 0, x.shape[0] - 1)

        # do polyfit on unique sized intervals (performance hack allowing optimisation)
        size = b1 - b0
        for s in np.unique(size):
            msk = (size == s)
            bnds = np.array(
                [(x[b0[i]], x[b1[i]]) for i in range(len(b0)) if msk[i]])  # get wavelength range of each interval
            XX = np.array([X[i, b0[i]:b1[i]+1] for i in range(len(b0)) if msk[i]])  # get spectral data
            xx = np.linspace(0, 1, s+1)  # use this as our x-variable so we can fit multiple regressions in one hit!

            # fit regressions
            _a, _b, _c = np.polyfit(xx, XX.T, deg=2)

            # compute turning point (position in x and height)
            p = (-_b / (2 * _a)) * ((bnds[:, 1] - bnds[:, 0])) + bnds[:,
                                                                 0]  # refine position [ and scale it back to wavelength range ]
            h = _c - (_b ** 2 / (4 * _a))  # refine height [ n.b. this doesn't need to be scaled!]

            # store
            x1[msk, j - 1] = h
            x1[msk, j] = np.clip(p, bnds[:,0], bnds[:,1]) # clip to possible domain in case of bad fitting
            x1[msk, j - 1][_a > 0] = 0  # this was a minima not a maxima!
    return x1

def minimum_wavelength(data, minw, maxw, method='gaussian', trend='hull', n=1, log=False,
                       sym=False, minima=True, k=4, hthresh=0.025, nthreads=1, vb=True, **kwds):
    """
    Perform minimum wavelength mapping to map the position of absorbtion features.

    Args:
        data: the hyperspectral data (e.g. image or point cloud) to perform minimum wavelength mapping on.
        minw: the lower limit of the range of wavelengths to search (in nanometers).
        maxw: the upper limit of the range of wavelengths to search (in nanometers).
        method: the method/model used to quantify the feature. Options are:

         - "minmax" - Identifies the n most prominent local minima to approximately resolve absorbtion feature positions. Fast but inaccurate.
         - "quad" - Applies the quadratic method to interpolate absorption depth using 3-bands adjacent to the local minima. See https://doi.org/10.1016/j.rse.2011.11.025 .
         - "poly" - Fits a quadratic to k bands on either side of the local minima. Similar to "quad" but with more than 3 points constraining the fitting. Hence, slower than quad but faster than gauss.
         - "gaussian" - fits n gaussian absorbtion features to the detrended spectra. Slow but most accurate.

        trend: the method used to detrend the spectra. Can be 'hull' or None. Default is 'hull'.
        log: If True, absorptions are modelled to log-spectra rather than inverted spectra.
        n: the number of features to fit. Note that this is not compatible with method = 'tpt'.
        minima: True if features should fit minima. If False then 'maximum wavelength mapping' is performed.
        sym: True if symmetric gaussian fitting should be used. Default is False.
        k: the number of adjacent measurements to look at during detection of local minima. Default is 4. Larger numbers ignore smaller features as noise.
        nthreads: the number of threads to use for the computation. Default is 1 (no multithreading).
        hthresh: the minimum feature depth before we bother fitting gaussians. Avoids wasting time / effort on flat spectra.
        vb: True if graphical progress updates should be created.
        **kwds: Keywords are passed to gfit.gfit( ... ).

    Returns:
        A MWL (n>1) or HyData instance containing the minimum wavelength mapping results.
    """

    # get relevant bands and detrend
    hc = data.export_bands((minw, maxw))
    if not minima: # flip if need be
        hc.data = np.nanmax(hc.data)-hc.data
    if trend is not None:
        if 'hull' in trend:
            hc = get_hull_corrected(hc,vb=vb)
        else:
            assert False, "Error - %s is an unsupported detrending method." % trend

    # setup output
    if sym:
        out = np.full((np.prod(data.data.shape[:-1]), n * 3), np.nan)
    else:
        out = np.full((np.prod(data.data.shape[:-1]), n * 4), np.nan)

    # remove invalid y-values from input
    S = hc.X()
    mask = np.isfinite(S).all(axis=-1)  # drop nans
    mask = mask & (S != S[:, 0][:, None]).any(axis=1)  # drop flat spectra (e.g. all zeros)
    X = S[mask]

    if mask.any():  # if no valid spectra, skip to end

        # flip hull corrected spectra as gfit only fits maxima
        if log:
            X = -np.log(X)
        else:
            X = 1 - X

        # get initial values
        x = hc.get_wavelengths()
        x0 = initialise(x, X, n, sym=sym, d=k, nthreads=nthreads)  # get initial values [ minmax method ]
        if 'minmax' in method:
            out[mask, :] = x0  # just use x0 as output
        elif 'gauss' in method:
            # drop low-amplitude features
            mask2 = x0[:,0] > hthresh # N.B. x0[:,0] will contain the deepest depth of each spectra

            # fit gaussians (and update x0 to make masking / reshaping easier)
            x0[ mask2, : ]= gfit(x, X[mask2,:], x0[mask2,:], n, sym=sym, nthreads=nthreads, vb=vb, **kwds)  # fit gaussians

            # store
            out[mask, :] = x0
        elif 'quad' in method:
            ii = np.arange(X.shape[0])
            for i in range(1, x0.shape[-1], 4):
                ix1 = np.array([hc.get_band_index(b) for b in x0[..., i]])
                ix0 = np.clip(ix1 - 1, 0, hc.band_count() - 1)
                ix2 = np.clip(ix1 + 1, 0, hc.band_count() - 1)

                # do some funky math (cf. https://doi.org/10.1016/j.rse.2011.11.025 )
                an = (X[ii, ix1] - X[ii, ix0]) * (x[ix0] - x[ix2]) + (X[ii, ix2] - X[ii, ix1]) * (x[ix1] - x[ix0])
                ad = (x[ix0] - x[ix2]) * (x[ix1] ** 2 - x[ix0] ** 2) + (x[ix1] - x[ix0]) * (x[ix2] ** 2 - x[ix0] ** 2)
                a = an / ad
                b = ((X[ii, ix1] - X[ii, ix0]) - a * (x[ix1] ** 2 - x[ix0] ** 2)) / (x[ix1] - x[ix0])

                m = -b / (2 * a)  # this is our adjusted minima :-)
                out[mask, :] = x0 # use minmax method for most things
                out[mask, i ] = m # update position
        elif 'poly' in method:
            out[mask, :] = _refine(x, X, x0, n=k, vb=vb)
        else:
            assert False, "Error - %s is an unsupported fitting method." % method

    # reshape outputs and add to HyData instance
    mwld = data.copy(data=False)
    mwld.data = out.reshape(data.data.shape[:-1] + (out.shape[-1],))

    # setup mwl collection and return
    mwl = MWL('M', '')
    mwl.bind( mwld, n, x=hc.get_wavelengths(), sym=sym, X=hc )
    return mwl


class mwl_legend(object):
    """
    A utility class storing data needed to create a legend for mwl plots.
    """

    def __init__(self, minh, maxh, minc, maxc, mode='val', cmap='rainbow', **kwds):
        """
        Create an mwl_legend instance.

        Args:
            minh: the value (wavelength) mapped to hue of 0
            maxh: the value (wavelength) mapped to a hue of 1
            minc: the value (typically depth/strength) mapped to brightness/saturation of 0.
            maxc: the value (typically depth/strength) mapped to brightness/saturation fo 1.
            mode: specifies if minc and maxc refer to brightness ('val', default) or saturation ('sat').
            cmap: the colormapping to use to determine hue.
            **kwds: Keywords can include:

                - xlab = a custom name for the hue (x) label. Default is 'Wavelength (nm)'
                - ylab = a custom name for the second axis. Default is 'Strength'.
        """

        self.minh = minh
        self.maxh = maxh
        self.minc = minc
        self.maxc = maxc
        self.xlab = kwds.get("xlab", 'Wavelength (nm)')
        self.ylab = kwds.get("ylab", 'Strength')
        self.mode = mode
        self.cmap = cmap

    def quick_plot(self, path : str = '', dpi : int = 300 ):
        """
        Quickly plot this MWL legend.

        Args:
            path: a path to save the legend to (as a .png). If None (default) then the file will not be written.
            dpi: the DPI to save the legend with
        Returns:
            The matplotlib figure and axes that the plot was drawn to.
        """

        fig, ax = plt.subplots(figsize=(7, 2))
        self.plot(ax, pos=(0.1, 0.1), s=(0.8, 0.8))
        ax.axis('off')
        if path != '':
            fig.savefig( path, dpi=dpi)
        return fig, ax

    def plot(self, ax, pos='top left', s=(0.3, 0.2)):
        """
        Add this legend to the specified figure.

        Args:
            ax: the axes to overlay the legend on.
            pos: the position of the legend. Can be:

                - a string: 'top left' (default), 'bottom left', 'top right' or 'bottom right', or;
                - a tuple (x,y) coordinates of the top left of the legend in figure coordinates.
            s: the (width, height) of the legend in figure coordinates. Default is (0.2, 0.1).
        Returns:
            the axis that was added to the plot.
        """

        # create axis
        pad = 0.025
        tickLeft = True
        tickBottom = True
        origin = 'upper'
        if isinstance(pos, str):
            if 'top' in pos.lower():
                y0 = 1.0 - s[1] - pad
            elif 'bottom' in pos.lower() or 'lower' in pos.lower():
                y0 = pad
                tickBottom = False
                origin = 'lower'
            else:
                assert False, "Error - unknown position %s" % pos

            if 'left' in pos.lower():
                x0 = pad
                tickLeft = False
            elif 'right' in pos.lower():
                x0 = 1.0 - s[0] - pad
            elif 'middle' in pos.lower() or 'centre' in pos.lower():
                x0 = 0.5 - (s[0] * 0.5)
            else:
                assert False, "Error - unknown position %s" % pos
        else:
            x0, y0 = pos

        ax = ax.inset_axes([x0, y0, s[0], s[1]])
        if not tickLeft:
            ax.yaxis.tick_right()
            ax.yaxis.set_label_position("right")
        if not tickBottom:
            ax.xaxis.tick_top()
            ax.xaxis.set_label_position("top")

        # fill axes using imshow
        if 'swir' in str(self.cmap):  # again - this is dirty dirty hack...
            extent = (2150.0, 2380.0, self.minc, self.maxc)

            # calculate hue
            h = np.linspace(2150., 2380., num=int(2000 * s[0]))
            _x = np.array([0, 2150,
                           2175, 2190, 2220,  # white mica
                           2245, 2254, 2261,  # chlorite/biotite/epidote
                           2325, 2330, 2345,  # carbonate
                           2360, 3000])

            _y = np.array([0.1, 0.1,
                           0.15, 0.3, 0.44,  # white mica
                           0.48, 0.5, 0.54,  # chlorite/biotite/epidote
                           0.82, 0.85, 0.94,  # carbonate
                           1.0, 1.0])
            wav = np.linspace(2150.0, 2380.0, 255)
            lookup = np.interp(wav, _x, 1 - _y)
            idx = (((h - 2150.0) / (2380.0 - 2150.0)) * lookup.shape[0]).astype(int)
            idx[idx < 0] = 0
            idx[idx > 254] = 254
            x = lookup[idx]
        elif 'rainbow' in self.cmap:
            extent = (self.minh, self.maxh, self.minc, self.maxc)
            x = np.linspace(0, 1, num=int(2000 * s[0]))
        else:
            extent = (self.minh, self.maxh, self.minc, self.maxc)
            x = np.linspace(0, 1, num=int(2000 * s[0]))
            cm = mpl.cm.get_cmap(self.cmap)
            rgba = cm(x)  # compute color from cmap
            x = mpl.colors.rgb_to_hsv(rgba[:, :3])[:, 0]  # update hue value accordingly

        # build image grid
        y = np.linspace(0, 1, num=int(2000 * s[1]))
        xx, yy = np.meshgrid(x, y)
        zz = np.full(xx.shape, 0.8)
        if 'val' in self.mode.lower():
            ax.imshow(matplotlib.colors.hsv_to_rgb(np.dstack([xx, zz, yy])),
                      origin=origin, aspect='auto', extent=extent)
        else:
            assert 'sat' in self.mode.lower(), "Error - %s is an invalid mode. Should be 'sat' or 'val'." % self.mode
            ax.imshow(matplotlib.colors.hsv_to_rgb(np.dstack([xx, yy, zz])),
                      origin=origin, aspect='auto', extent=extent)

        ax.set_xlabel(self.xlab)
        ax.set_ylabel(self.ylab)
        ax.set_yticks([])
        return ax


def colourise_mwl(mwl, mode='p-d', cmap='rainbow', **kwds):
    """
    Takes a HyData instance containing minimum wavelength bands (pos, depth, width and strength) and creates
    a RGB composite such that hue ~ pos, sat ~ width and val ~ strength or depth.

    Args:
        mwl: the HyData instance containing minimum wavelength data.
        mode: the mapping from position (p), width (w) and depth and (d) to hsv. Default is 'pwd', though other options
              are 'p-d' (constant saturation of 80%), 'pdw' and 'pd-' (constant brightness of 85%).
        cmap: the colour mapping to use. Default is to map position to hue ('rainbow'), but any matplotlib
              colormap string can be provided here. Note that colours output by this colormap will be scaled in brightness
              and saturation according to the mode argument. Alternatively use 'swir' for customised
              rainbow-like colour stretch optimised for clay, mica and chlorite absorbtion features in the SWIR.
        **kwds: Keywords can include:

              - hue_map = Wavelengths (xxx.x, yyy.y) or percentiles (x,y) to use when converting wavelength to hue.
                            Default is (0,100).
              - sat_map = Widths (xxx.x, yyy.y) or percentiles (x,y) to use when converting width to saturation.
                            Default is (0,100).
              - val_map = Strengths/depths (xxx.x, yyy.y) or percentiles (x,y) to use when converting strength to brightness.
                            Default is (0,75), as these tend to be dominated by low/zero values.

    Returns:
        A tuple containing:

        - either an RGB HyImage object (if mwl is an image) or the original HyCloud with defined rgb bands.
        - cmap = a mwl_legend instance for plotting colour maps.
    """

    # extract data
    assert (mwl.band_count() == 4) or (
                mwl.band_count() == 3), "Error - HyData instance does not contain minimum wavelength data?"

    h = mwl.X()[..., 1].copy()  # pos
    s = mwl.X()[..., 2].copy()  # width
    v = mwl.X()[..., 0].copy()  # depth

    # normalise to range 0 - 1
    stretch = ['hue_map', 'sat_map', 'val_map']
    ranges = []  # store data ranges for colour map
    for i, b in enumerate([h, s, v]):
        if 'swir' in str(kwds.get(stretch[i], '')).lower():
            ranges.append((2150.0, 2380.0))
            continue

        mn, mx = kwds.get(stretch[i], (0, 100))
        if i == 2:  # use different default stretch for value (as these tend to be heavily skewed)
            mn, mx = kwds.get(stretch[i], (0, 75))

        if isinstance(mn, int):  # convert percentiles to values
            mn = np.nanpercentile(b, mn)
        if isinstance(mx, int):  # convert percentiles to values
            mx = np.nanpercentile(b, mx)
        ranges.append((mn, mx))  # store data ranges for colour map

        # apply stretch
        b -= mn
        b /= (mx - mn)
        b[b < 0] = 0
        b[b > 1] = 1

    # flip width (so low width = high saturation)
    s = 1 - s

    # special case - width is all the same
    if not np.isfinite(s).any():
        s = np.full(s.shape[0], 0.8)  # fill mostly saturated

    # remove nans
    mask = np.logical_not(np.isfinite(h) & np.isfinite(s) & np.isfinite(v))
    h[mask] = 0
    v[mask] = 0
    s[mask] = 0

    # calculate hue value from colormap
    if 'rainbow' in cmap.lower():
        h = h  # no change to hue
    elif 'swir' in cmap.lower():  # use SWIR customised colormap
        _x = np.array([0, 2150,
                       2175, 2190, 2220,  # white mica
                       2245, 2254, 2261,  # chlorite/biotite/epidote
                       2325, 2330, 2345,  # carbonate
                       2360, 3000])

        _y = np.array([0.1, 0.1,
                       0.15, 0.3, 0.44,  # white mica
                       0.48, 0.5, 0.54,  # chlorite/biotite/epidote
                       0.82, 0.85, 0.94,  # carbonate
                       1.0, 1.0])
        wav = np.linspace(2150.0, 2380.0, 255)
        lookup = np.interp(wav, _x, 1 - _y)
        idx = (((h - 2150.0) / (2380.0 - 2150.0)) * lookup.shape[0]).astype(int)
        idx[idx < 0] = 0
        idx[idx > 254] = 254
        h = lookup[idx]
    else:  # use matplotlib colormap
        cm = mpl.cm.get_cmap(cmap)
        rgba = cm(h)
        h = mpl.colors.rgb_to_hsv(rgba[:, :3])[:, 0]  # update hue value accordingly

    # convert to rgb based on mapping mode
    if 'pwd' in mode.lower():  # pos, width, depth (default)
        rgb = matplotlib.colors.hsv_to_rgb(np.array([h, s, v]).T)
    elif 'pdw' in mode.lower():  # pos, depth, width
        s[mask] = 0.5
        rgb = matplotlib.colors.hsv_to_rgb(np.array([h, v, s]).T)
    elif 'p-d' in mode.lower():  # pos, const, depth
        rgb = matplotlib.colors.hsv_to_rgb(np.array([h, np.full(len(h), 0.8), v]).T)
    elif 'pd-' in mode.lower():  # pos, depth, const
        rgb = matplotlib.colors.hsv_to_rgb(np.array([h, v, np.full(len(h), 0.8)]).T)

    # add nans back in
    rgb[mask, :] = np.nan

    # create a colourbar object
    if 'pdw' in mode.lower() or 'pd-' in mode.lower():
        cbar = mwl_legend(ranges[0][0], ranges[0][1],
                          ranges[2][0], ranges[2][1], mode='sat', cmap=cmap)
    else:
        cbar = mwl_legend(ranges[0][0], ranges[0][1],
                          ranges[2][0], ranges[2][1], mode='val', cmap=cmap)
    if 'swir' in str(kwds.get('hue_map', '')).lower():
        cbar.minh = 'swir'
        cbar.maxh = 'swir'

    if mwl.is_image():
        out = mwl.copy(data=False)
        out.data = rgb.reshape((mwl.data.shape[0], mwl.data.shape[1], 3))
        out.set_band_names([0, 1, 2])
        out.set_wavelengths(None)
        return out, cbar
    else:  # HyCloud - set rgb and return
        # store
        out = mwl.copy(data=False)
        out.data = rgb
        out.set_band_names(["r", "g", "b"])

        # set rgb
        rgb[rgb <= 0.01] = 0.01  # avoid 0 as these are masked by plotting functions
        out.rgb = (rgb * 255).astype(np.uint8)
        return out, cbar


def plot_ternary(F1, F2, F3, bounds, weights=[1., 1., 1.], subsample=1, depth_thresh=0.01, ax=None, **kwds):
    """
    Plot a ternary diagram comparing the depths and positions of three minimum wavelength features.

    Args:
        F1 (HyData): A HyData instance containing feature depth and position as band 0 and 1 respectively.
        F2 (HyData): A HyData instance containing feature depth and position as band 0 and 1 respectively.
        F3 (HyData): A HyData instance containing feature depth and position as band 0 and 1 respectively.
        bounds: List containing min and max position values for each feature [ (min,max), (min,max), (min,max)].
        weights: weights applied to each depth. Default is [1,1,1].
        subsample: skip entries in the data features for large datasets. Default is 1 (don't skip any).
        depth_thresh: the minimum depth to be considered a valid feature. Default is 0.01.
        ax: an axes to plot to. If None (default) then a new axes is created.
        **kwds: Keywords can include:

             - r = the diameter of the ternary diagram (default is 0.8)
             - figsize = the figure size as a (width,height) tuple. Default is (10,10)
             - w = the width of edge plots. Default is 0.2.
             - gs = the number of grid lines. Default is 5.
             - invalid = invalid region (don't plot position if abundance < this value). Default is 0.2.
             - title = the title of the plot.
             - labels = the names of each of the features (F1, F2 and F3) used for labelling.
             - label_offset = the space between each label and the relevant vertex.
             - s = the point size. Default is 4.
             - a = point alpha. Default is 0.1.
    """
    # put features in a list
    features = [F1, F2, F3]

    # get axes
    if ax is None:
        fig, ax = plt.subplots(figsize=kwds.get('figsize', (10, 10)))
    else:
        fig = ax.get_figure()
    ax.set_aspect('equal')

    # compute coordinates of border
    r = kwds.get('r', 0.8)  # diameter of ternary diagram
    w = kwds.get('w', 0.2)  # width of edge plots
    gs = kwds.get('gs', 5)  # number of grid points
    invalid = kwds.get('invalid', 0.2)  # invalid region (don't plot position if abundance < this value)
    title = kwds.get('title', 'Absorption features')
    label_offset = kwds.get('label_offset', 0.2)
    labels = kwds.get('labels', ['F1', 'F2', 'F3'])
    edge_bounds = bounds
    colors = ['r', 'g', 'b']  # colours for the labels. These should not be changed.
    psize = kwds.get('s', 4)
    palpha = kwds.get('a', 0.1)

    # plot triangle and outline of ternary diagram
    X = np.array([[0, np.cos(np.deg2rad(30)), -np.cos(np.deg2rad(30))],
                  [1, -np.sin(np.deg2rad(30)), -np.sin(np.deg2rad(30))]]).T
    X *= r  # apply scale factor

    def baryToCC(a, b, c):
        # normalise
        sm = a + b + c
        for v in [a, b, c]:
            v /= sm

        # compute coordinates
        C, A, B = X
        x = A[0] * a + B[0] * b + C[0] * c
        y = A[1] * a + B[1] * b + C[1] * c
        return x, y

    def edgeToCC(edge, abundance, y):
        C, A, B = X
        if 'a' in edge.lower():
            xx = A - C
            o = C  # edge a starts at C and ends at A
        if 'b' in edge.lower():
            xx = B - A
            o = A  # edge b starts at A and ends at B
        if 'c' in edge.lower():
            xx = C - B
            o = B  # edge c starts at B and ends at C

        yy = xx[::-1] * np.array([-1, 1])  # y is perpendicular to x
        yy = yy * np.linalg.norm(yy) * w  # normalised to correct length

        x = o[0] + xx[0] * abundance + yy[0] * y
        y = o[1] + xx[1] * abundance + yy[1] * y
        return x, y

    # plot triangle
    ax.plot(X[[0, 1, 2, 0], 0], X[[0, 1, 2, 0], 1], color='k', lw=2, zorder=1)

    # plot ternary grid
    for e, c in zip('abc', 'gbr'):
        V = []
        ee = 'abc'.replace(e, '')
        for a in np.linspace(0, 1, gs):
            x0, y0 = edgeToCC(ee[0], a, 0)
            x1, y1 = edgeToCC(ee[1], 1 - a, 0)

            # plot edge
            ax.plot([x0, x1], [y0, y1], color=c, zorder=0, alpha=0.4, ls=':')

    # plot edges
    for e, c in zip('abc', 'rgb'):
        ax.plot(*edgeToCC(e, np.array([0, 0, 1, 1]), np.array([0, 1, 1, 0])), color=c, lw=2)

    # plot edge grid
    for e, c in zip('abc', 'rgb'):
        for a in np.linspace(0, 1, gs):
            ax.plot(*edgeToCC(e, np.array([0, 1]), np.array([a, a])), color='k', zorder=0, alpha=0.4, ls=':')
            ax.plot(*edgeToCC(e, np.array([a, a]), np.array([0, 1])), color=c, zorder=0, alpha=0.4, ls=':')

    # plot invalid areas
    for e in 'abc':
        ax.fill(*edgeToCC(e, np.array([0, 0, invalid, invalid]), np.array([0, 1, 1, 0])), color='k', zorder=-1, lw=0,
                alpha=0.1)

    # plot component labels
    for i, e in enumerate('abc'):
        co = np.array(edgeToCC(e, 1, 0))
        a = np.zeros(3)
        a[i] = 1 - label_offset
        off = co - np.array(baryToCC(*a))
        ax.text(*(co + off), s=labels[i], fontsize='xx-large', ha='center', va='top', color=colors[i])

    # plot edge labels
    for i, (e, a) in enumerate(zip('abc', [-60, 0, 60])):
        mn = edgeToCC(e, 0.05, 0.15)
        ax.text(*mn, s=str(edge_bounds[i][0]), ha='center', va='center', rotation=a)
        mx = edgeToCC(e, 0.05, 0.85)
        ax.text(*mx, s=str(edge_bounds[i][1]), ha='center', va='center', rotation=a)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_axis_off()

    # plot data
    depth = np.nan_to_num(np.array([D.data[..., 0].ravel() for D in features]))
    depth *= np.array(weights)[:, None]
    depth /= np.sum(depth, axis=0)[None, :]  # normalise to sum to one
    pos = np.array([D.data[..., 1].ravel() for D in features])
    mask = (depth > depth_thresh).any(axis=0)  # don't plot points with only shallow features
    pos = pos[:, mask][:, ::subsample]
    depth = depth[:, mask][:, ::subsample]

    # compute colours
    rgb = (depth / np.sum(depth, axis=0)).T
    ax.scatter(*baryToCC(depth[0, :], depth[1, :], depth[2, :]),
               color=rgb, alpha=palpha, s=psize)

    for i, e in enumerate('abc'):
        y = pos[i, :] - edge_bounds[i][0]
        y = np.clip(y / (edge_bounds[i][1] - edge_bounds[i][0]), 0, 1)
        mask = depth[i, :] > invalid
        ax.scatter(*edgeToCC(e, depth[i, mask], y[mask]),
                   color=rgb[mask], alpha=palpha, s=psize)

    ax.set_title(title, size='xx-large')
    return fig, ax


"""
A collection of functions for performing histogram equalisation and other similar colour-matching routines.
"""

import numpy as np

def hist_eq(adj, ref):
    """
    Perform histogram normalisation on two datasets, as describe at: https://stackoverflow.com/questions/32655686/histogram-matching-of-two-images-in-python-2-x.

    Args:
        adj: the source data to transform
        ref: data following the distribution to be matched too.
    Returns:
        an array of the same shape as source that contains the transformed data.
    """
    olddtype = adj.dtype
    oldshape = adj.shape
    adj = adj.ravel()
    ref = ref.ravel()
    src_msk = np.isfinite(adj)
    tmp_msk = np.isfinite(ref)

    if np.isfinite(adj).any() and np.isfinite(ref).any():  # we need at least some finite values
        s_values, bin_idx, s_counts = np.unique(adj[src_msk], return_inverse=True,
                                                return_counts=True)
        t_values, t_counts = np.unique(ref[tmp_msk], return_counts=True)
        s_quantiles = np.cumsum(s_counts).astype(np.float64)
        s_quantiles /= s_quantiles[-1]
        t_quantiles = np.cumsum(t_counts).astype(np.float64)
        t_quantiles /= t_quantiles[-1]
        interp_t_values = np.interp(s_quantiles, t_quantiles, t_values)
        interp_t_values = interp_t_values.astype(olddtype)

        # update output and return
        adj[src_msk] = interp_t_values[bin_idx]

    return adj.reshape(oldshape)


def norm_eq(adj, adj_s, ref_s, per_band=True, inplace=False):
    """
    Match image colour by assuming normally distributed data, and so
    transform adj so it has the same mean and standard deviation as ref.

    (loosely based on https://www.cs.tau.ac.il/~turkel/imagepapers/ColorTransfer.pdf )

    Args:
        adj: the source data to transform.
        adj_s: subset of adj to calculate statistics from (e.g. matching pixels).
        ref_s: reference subset to match statistics too (e.g. matching pixels from reference image).
        per_band: True (default) if the correction should be applied to each band separately. This can
                  potentially introduce spectral artefacts so use with care. If False, the whole dataset is
                  scaled/recentered at once, so the spectra will not be distorted.
        inplace: True if data should be modified in place rather than being copied. Default is False.
    """

    assert isinstance(adj, np.ndarray), "Adj must be numpy array."
    assert isinstance(adj_s, np.ndarray), "Ref must be numpy array."
    assert isinstance(ref_s, np.ndarray), "Ref must be numpy array."

    # copy?
    if not inplace:
        adj = adj.copy()

    # calculate means and standard deviation of each channel
    if per_band:
        a = tuple(range(len(adj_s.shape) - 1))  # apply per band
    else:
        a = tuple(range(len(adj_s.shape)))  # apply to whole dataset

    Mu1 = np.nanmean(adj_s, axis=a)
    Mu2 = np.nanmean(ref_s, axis=a)
    Sd1 = np.nanstd(adj_s, axis=a)
    Sd2 = np.nanstd(ref_s, axis=a)

    adj -= Mu1  # center on mean
    adj *= Sd2 / Sd1  # match standard deviations (basically adjust contrast)
    adj += Mu2  # recenter onto target mean

    return adj


import os
import hylite.io as io
from .sensor import Sensor
import numpy as np
from pathlib import Path
import cv2
from multiprocessing import Pool
from tqdm import tqdm
import piexif
import matplotlib.pyplot as plt

class Rikola(Sensor):
    """
    Implementation of sensor corrections for the Rikola sensor.
    """

    # constants used for boundary tracking when correcting for sensor shift (spatial)
    LEFT_BOUND = 99999
    RIGHT_BOUND = 99998
    TOP_BOUND = 99997
    BOTTOM_BOUND = 99996

    @classmethod
    def correct_image(cls, image, verbose=True, **kwds):
        """
        Apply sensor corrections to an image.

        Args:
            image (hylite.HyImage): a hyImage instance of an image captured using this sensor.
            verbose (str): true if updates/progress should be printed to the console. Default is True.
            **kwds: Optional keywords include:

                 - crop = True if the image should be cropped so that only pixels covered by all bands after alignment are retained. Default is True.
                 - align = True if the image bands should be coaligned using sift (due to delayed aquisition times). Default is True.
                 - lens = True if the RIKOLA lens correction should be apply to correct for lens distortion. Default is True.
                 - match_band = the band to co-register data to. This should be a good quality band (low noise etc.). Default is 30.
                 - contrast_thresh = the contrast threshold used for SIFT matching. Default is 0.01.
                 - sigma = the sigma to use for SIFT matching. Default is 1.5.
                 - edge_thresh = the edge threshold to use for sift matching. Default is 10.
                 - eq = equalise before sift? 'True' or 'False'. Default is 'False'.
                 - match_dist = maximum matching distance for SIFT. Default is 0.7.
                 - min_match = the minimum number of matches to apply band coregistration. Default is 5.
                 - warp = the warping method used to align bands. Can be 'affine', 'homography' or 'flow'.
                   Default is 'flow' aligns each band using 'affine' and then refines the result using dense optical flow.
                   This is slow but corrects for errors associated with perspective and non-flat topography....

        """

        #get keywords
        contrast_thresh = kwds.get("contrast_thresh", 0.01)
        sigma = kwds.get("sigma", 1.5)
        edge_thresh = kwds.get("edge_thresh", 10)
        eq = kwds.get("eq", False)
        match_dist = kwds.get("match_dist", 0.7)
        min_match = kwds.get("min_match", 5)
        warp = kwds.get("warp", 'flow')
        match_band = kwds.get("match_band", 30)
        crop = kwds.get('crop', True)
        align = kwds.get('align', True)
        lens = kwds.get('lens', True)

        # n.b. rikola camera already applies dark correction. Hence, we only need apply a lens correction and align the bands

        ###############################
        # LENS CORRECTION
        ###############################

        #Lens correction for NEW rikola camera
        if image.samples() == 1024: #new rikola

            #delete dodgy bands at start and end of spectral range
            image.band_names = image.header.get_wavelengths() #np.fromstring( image.header['wavelength'], sep="," ) #load band wavelengths
            lower = np.argmin( image.band_names < 512 ) #first valid band
            upper = np.argmin( image.band_names < 914 ) #last valid band
            image.data = image.data[:,:,lower:upper] #delete dodgy bands
            image.band_names = image.band_names[lower:upper]

            if lens:
                #lens correction for sensor 1
                RIKMat = np.array(
                    [[1559.169765826131 - 2.1075939987083148, -0.46558291934094703, 512. - 12.049303428088125],
                     [0, 1559.169765826131, 512. + 10.043851873473248], [0, 0, 1]])
                RIKdist = np.array([[-0.22431431139011579, -0.25788278936781356, -0.00070123282637077225,
                                      0.0023607284552227088, 1.3341497419995652]])
                m, roi = cv2.getOptimalNewCameraMatrix(RIKMat, RIKdist, (image.lines(), image.samples()), 1,
                                                       (image.lines(), image.samples()))


                #lens correction for sensor 2
                RIKMat2 = np.array(
                    [[1562.7573572611377 - 3.2133196257811703, 1.6878264077638325, 512. - 14.06301214860064],
                     [0, 1562.7573572611377, 512. + 11.549906117342049], [0, 0, 1]])
                RIKdist2 = np.array([[-0.23799506683807267, -0.079495552793610855, 0.00085334990816270302,
                                      0.0017596141854012402, 0.78590417508343857]])
                m2, roi2 = cv2.getOptimalNewCameraMatrix(RIKMat2, RIKdist2, (image.lines(), image.samples()), 1,
                                                       (image.lines(), image.samples()))

                #find sensor gap where we change corrections (648 nm band)
                sensgap = np.argmin(image.get_wavelengths() < 648)

                # transpose into img[y][x][b]
                image.data = np.transpose(image.data, (1, 0, 2))

                # apply lens correction
                if verbose: print("Applying lens correction... ", end="", flush="True")
                image.data[:,:,0:sensgap] = cv2.undistort(image.data[:,:,0:sensgap], RIKMat, RIKdist, None, m)
                image.data[:,:,sensgap:] = cv2.undistort(image.data[:, :, sensgap::], RIKMat2, RIKdist2, None, m2)

                # crop
                x, y, w, h = roi2
                image.data = image.data[x:(x + w - 10), y:(y + h - 10)]

                if verbose: print("DONE.", flush="True")

                # transpose back to img[x][y][b]
                image.data = np.transpose(image.data, (1, 0, 2))


        #lens correction for OLD rikola cameara
        elif lens:
            # define camera correction parameters
            # camera matrix = ([[fx,skew,cx],[0,fy,cy],[0,0,1]])
            # distortion coefficients = ([[k1,k2,kp1,p2,k3]])
            # n.b. depending on the aquisition mode the rikola reads either the whole sensor (1010 lines) or half the sensor (658 lines). Hence we chose the camera matrix accordingly
            if image.lines() == 1010: #old rikola, full frame
                RIKMat = np.array([[1580, -0.37740987561726941, 532.14269389794072],
                                   [0, 1586.5023476977308, 552.87899983359080],
                                   [0, 0, 1]])

                RIKdist = np.array([[-0.34016504377397944, 0.15595251253428483, 0.00032919179911211665,
                                     0.00016579344155373088, 0.051315602289989909]])

            elif image.lines() == 648: #old rikola, half frame
                RIKMat = np.array([[1580.9821817891338, -0.053468464819987738, 537.09531859948970],
                                   [0, 1580.4094746112266, 369.76442407125506],
                                   [0, 0, 1]])
                RIKdist = np.array([[-0.31408677145500508, -0.26653256083139154, 0.00028155583639827883,
                                     0.00025705469073531660, 2.4100464839836362]])
            else:
                assert False, "Error - invalid number of lines (%d) for RIKOLA image. Should be either 1010 or 648" % (
                    image.lines())

            # create calibration map
            m, roi = cv2.getOptimalNewCameraMatrix(RIKMat, RIKdist, (image.lines(), image.samples()), 1,
                                                   (image.lines(), image.samples()))

            # transpose into img[y][x][b]
            image.data = np.transpose(image.data, (1, 0, 2))

            # loop through bands and apply correction
            if verbose: print("Applying lens correction... ", end="", flush="True")

            # apply lens correction
            image.data = cv2.undistort(image.data, RIKMat, RIKdist, None, m)

            # crop
            x, y, w, h = roi
            image.data = image.data[x:(x + w - 10), y:(y + h - 10)]

            if verbose: print("DONE.", flush="True")

            # transpose back to img[x][y][b]
            image.data = np.transpose(image.data, (1, 0, 2))

        ###############################
        # COREGISTER BANDS
        ###############################

        # flag boundary pixels in each band so we can crop it after warping
        image.edge_mask = np.zeros(image.data[:, :, 0:4].shape, dtype=np.float32)
        image.edge_mask[0:3, :, 0] = 1.0
        image.edge_mask[-4:-1, :, 1] = 1.0
        image.edge_mask[:, 0:3, 2] = 1.0
        image.edge_mask[:, -4:-1, 3] = 1.0
        edge_accum = np.zeros_like(image.edge_mask)
        if align:
            # identify SIFT features for each band
            if verbose: print("Identifying SIFT features... ", end="", flush="True")
            features = [image.get_keypoints(b) for b in range(image.band_count())]

            if verbose: print("DONE.", flush="True")

            if verbose: print("Stacking bands... ", end="", flush="True")

            # loop through bands and find matches
            matches = []
            for b in range(0, image.band_count()):
                if b == match_band:
                    matches.append(None)
                    continue  # skip reference band as it doesn't need to be moved...
                if verbose: print("Matching bands... %d " % int((100 * b) / image.band_count()) + "% \r", end="", flush="True")

                # can we match this band with the reference band?
                src_pts, dst_pts = io.HyImage.match_keypoints(features[b][0], features[match_band][0],
                                                         features[b][1], features[match_band][1],
                                                         method='SIFT', dist=match_dist)
                matches.append((src_pts, dst_pts))

            if verbose: print("Matching bands... DONE.                          ", flush="True")

            # apply transformation to each band
            for b in range(0, image.band_count()):
                if b == match_band: continue  # skip reference band as it doesn't need to be moved...
                if verbose: print("Stacking bands... %d " % int((100 * b) / image.band_count()) + "% \r", end="",flush="True")

                # get match points
                src_pts, dst_pts = matches[b]
                if src_pts is None:  # couldn't match with reference band, try matching with one of the already aligned bands
                    print("Warning - could not directly band %d. Attempting indirect alignment... " % b, end="",
                          flush="True")
                    i = 1
                    while src_pts is None:
                        if b - i < 0:  # run out of bands
                            print("Failed to align band" % b)
                            break

                        # calculate features in aligned band
                        feat2 = image.get_keypoints(b-i, eq=eq, contrastThreshold=contrast_thresh,
                                            edgeThreshold=edge_thresh, sigma=sigma)

                        # try matching
                        src_pts, dst_pts = io.HyImage.match_keypoints(features[b][0], feat2[0],
                                                                features[b][1], feat2[1],
                                                                dist=match_dist, method='sift')

                    print("Aligned to band %d" % (b - i))

                # do alignment
                if not src_pts is None:
                    if ('affine' in warp.lower()) or ('flow' in warp.lower()):
                        H, status = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0)
                        dst_mask = dst_pts[:, 0, :] * status
                        src_mask = src_pts[:, 0, :] * status
                        dst_mask = dst_mask[dst_mask.all(1)]
                        src_mask = src_mask[src_mask.all(1)]
                        dst_mask = np.expand_dims(dst_mask, axis=1)
                        src_mask = np.expand_dims(src_mask, axis=1)
                        #M = cv2.estimateRigidTransform(src_mask, dst_mask, False)
                        M = cv2.estimateAffinePartial2D(src_mask, dst_mask)[0]

                        # apply to image
                        image.data[:, :, b] = cv2.warpAffine(image.data[:, :, b], M,
                                                             (image.data.shape[1], image.data.shape[0]))

                        # apply to edge mask
                        for e in range(0, 4):
                            edge_accum[:, :, e] += cv2.warpAffine(image.edge_mask[:, :, e], M,
                                                                  (image.edge_mask.shape[1], image.edge_mask.shape[0]))

                    elif ('homography' in warp.lower()):
                        H, status = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 3.0)

                        # apply to image
                        image.data[:, :, b] = cv2.warpPerspective(image.data[:, :, b], H,
                                                                  (image.data.shape[1], image.data.shape[0]))

                        # apply to edge mask
                        for e in range(0, 4):
                            edge_accum[:, :, e] += cv2.warpPerspective(image.edge_mask[:, :, e], H,
                                                                       (image.edge_mask.shape[1],
                                                                        image.edge_mask.shape[0]))
                    else:
                        assert False, "Error - unknown transformation type. Should be 'flow' or 'homography' or 'affine'."

            # update edge mask
            image.edge_mask = edge_accum

            # crop image to area of complete overlap
            xmin = np.min(np.where(np.sum(image.edge_mask[:, :, 0], axis=1) == 0.0))
            xmax = np.max(np.where(np.sum(image.edge_mask[:, :, 1], axis=1) == 0.0))
            ymin = np.min(np.where(np.sum(image.edge_mask[:, :, 2], axis=0) == 0.0))
            ymax = np.max(np.where(np.sum(image.edge_mask[:, :, 3], axis=0) == 0.0))
            if crop:
                image.data = image.data[xmin:xmax, ymin:ymax, :]
            else:
                image.data[:xmin, :, :] = 0.
                image.data[xmax:, :, :] = 0.
                image.data[:, :ymin, :] = 0.
                image.data[:, ymax:, :] = 0.

            # apply flow transform to aligned bands
            if 'flow' in warp.lower():

                # reset edge mask
                image.edge_mask = np.zeros(image.data[:, :, 0:4].shape, dtype=np.float32)
                image.edge_mask[0:3, :, 0] = 1.0
                image.edge_mask[-4:-1, :, 1] = 1.0
                image.edge_mask[:, 0:3, 2] = 1.0
                image.edge_mask[:, -4:-1, 3] = 1.0
                edge_accum = np.zeros_like(image.edge_mask)
                alg = cv2.optflow.createOptFlow_DeepFlow()
                X, Y = np.meshgrid(range(image.data.shape[1]), range(image.data.shape[0]))
                map = np.dstack([X, Y]).astype(np.float32)
                bnd1 = np.uint8(255 * (image.data[:, :, 0] - np.nanmin(image.data[:, :, 0])) /
                         (np.nanmax(image.data[:, :, 0]) - np.nanmin(image.data[:, :, 0])))
                for b in range(1, image.band_count()):
                    if verbose: print("Warping bands... %d " % int((100 * b) / image.band_count()) + "% \r", end="",
                                      flush="True")
                    bnd2 = np.uint8(255 * (image.data[:, :, b] - np.nanmin(image.data[:, :, b])) /
                         (np.nanmax(image.data[:, :, b]) - np.nanmin(image.data[:, :, b])))
                    bnd1[bnd2 == 0.] = 0.
                    bnd2[bnd1 == 0.] = 0.
                    flow = alg.calc(bnd1, bnd2, None)
                    map[:, :, 0] += flow[:, :, 0]
                    map[:, :, 1] += flow[:, :, 1]
                    image.data[:, :, b] = cv2.remap(image.data[:, :, b], map, None, cv2.INTER_LINEAR)
                    image.data[:, :, b][bnd2 == 0.] = 0.
                    for e in range(0, 4):
                        edge_accum[:, :, e] += cv2.remap(image.edge_mask[:, :, e], map, None, cv2.INTER_LINEAR)
                    bnd1 = bnd2

                image.edge_mask = edge_accum

                # crop again

                xmin = np.min(np.where(np.sum(image.edge_mask[:, :, 0], axis=1) == 0.0))
                xmax = np.max(np.where(np.sum(image.edge_mask[:, :, 1], axis=1) == 0.0))
                ymin = np.min(np.where(np.sum(image.edge_mask[:, :, 2], axis=0) == 0.0))
                ymax = np.max(np.where(np.sum(image.edge_mask[:, :, 3], axis=0) == 0.0))
                if crop:
                    image.data = image.data[xmin:xmax, ymin:ymax, :]
                else:
                    image.data[:xmin, :, :] = 0.
                    image.data[xmax:, :, :] = 0.
                    image.data[:, :ymin, :] = 0.
                    image.data[:, ymax:, :] = 0.
            if verbose: print("Warping bands...  DONE.                         ", flush="True")

    @classmethod
    def correct_folder(cls, path, **kwds):
        """
        Many sensors use simple/common data structures to store data/headers/dark reference etc. Hence it is often easiest
        to pass an output folder to the sensor for correction.

        Args:
            path (str): a path to the folder containing the sensor specific data.
            **kwds: Keywords will be passed to correct_image, except for:

                - multi = True if multiple threads will be spawned to processes images in parallel. Default is True.
                - nthreads = the number of worker threads to spawn if multithreaded is true. Default is the number of CPU cores.

        Returns:
            A hyImage to which all sensor-specific corrections have been applied. Note that this will generally not include
            topographic or atmospheric corrections.
        """

        assert os.path.isdir(path), "Error - %s is not a valid directory path" % path
        assert os.path.exists(path), "Error - could not find %s" % path

        #get keywords
        multi = kwds.get('multi', True)
        nthread = kwds.get('nthreads', os.cpu_count())
        overwrite = kwds.get('overwrite', False)

        if nthread is None:
            print("Warning - could not identify CPU count. Multithreading disabled.")
            multi = False

        #find images to process
        images = list(Path(path).rglob("*Calib*[0-9][0-9][0-9][0-9][0-9].hdr")) #old rikola files
        images += list(Path(path).rglob("*Sequence*[0-9][0-9][0-9][0-9][0-9].hdr")) #new rikola files

        if not overwrite:  # ignore already corrected files
            images = [x for x in images if list(Path(path).rglob(x.name[:-4] + "_CORRECTED.hdr")) == []]

        corrected = []
        print("Processing %d images with the following settings: %s" % (len(images),kwds))
        if multi: #run in multithreaded
            p = Pool(processes = nthread) #setup pool object with n threads
            corrected = list(tqdm( p.imap_unordered(Rikola._cmp, [(i, kwds) for i in images]), total=len(images),leave=False)) #distribute tasks
        else: #not multithreaded
            for pth in tqdm(images,leave=False):
                # noinspection PyCallByClass
                corrected.append(Rikola._cmp( (pth, kwds) ) )#

        return corrected

    # define worker function for correct_folder( ... ) multithreading
    @classmethod
    def _cmp(cls, args):
        pth = str(args[0])
        kwds = args[1]
        image = io.load(pth)  # load image
        Rikola.correct_image(image, False, **kwds)  # correct image
        io.saveWithGDAL(os.path.splitext(pth)[0] + "_CORRECTED.hdr", image)  # save corrected image
        return os.path.splitext(pth)[0] + "_CORRECTED.hdr"  # return corrected path

    @classmethod
    def GPS_JPG(cls, MAIN):
        """
        Creates geotagged RGB JPGs for a folder of calibrated Rikola images and stores them in a subfolder "RGB/"

        Args:
            MAIN (str): path to the folder containing the image data (ending with "CORRECTED.dat")
                  as well as the acquisition-specific "TASKFILE.TXT"

        """

        try:
            import osgeo.gdal as gdal # todo - remove GDAL dependency for this function (replace with Pillow?)
        except:
            assert False, "Error - the GPS_JPG function requires GDAL to be installed."

        # create new folder for file storage
        if not os.path.exists(MAIN + 'RGB/'):
            os.makedirs(MAIN + 'RGB/')

        # read taskfile.txt containing GPS information
        with open(MAIN + 'TASKFILE.TXT') as f:
            task = f.readlines()
        task = np.asarray(task[60:])
        taskfile = np.asarray([task[i].split(",") for i in range(len(task))])

        # collect filelist for processing
        filelist = [root + '/' + file for root, _, files in os.walk(MAIN) for file in files if
                    "_CORRECTED.dat" in file and "xml" not in file]

        # processing for each file in filelist
        for file in filelist:

            # read 3 bands (RGB) as PIL image
            raster = gdal.Open(file)
            band1 = raster.GetRasterBand(23).ReadAsArray()
            band2 = raster.GetRasterBand(7).ReadAsArray()
            band3 = raster.GetRasterBand(1).ReadAsArray()
            arr1 = band1.astype(np.float32)
            maxim = np.nanmax(band1)
            band1_255 = np.uint8(arr1 / maxim * 255)
            arr1 = band2.astype(np.float32)
            maxim = np.nanmax(arr1)
            band2_255 = np.uint8(arr1 / maxim * 255)
            arr1 = band3.astype(np.float32)
            maxim = np.nanmax(arr1)
            band3_255 = np.uint8(arr1 / maxim * 255)

            # find and extract file-relevant info from taskfile
            rawfilename = os.path.split(file)[1].replace('_CORRECTED.dat', '.DAT')[6:]
            filenumber = np.where(taskfile[:, 0] == rawfilename)
            test = taskfile[filenumber, 16].tolist()
            altitude = float('[]'.join(test[0]))
            longi = taskfile[filenumber, 11].tolist()
            long = float('[]'.join(longi[0])) / 100
            lati = taskfile[filenumber, 9].tolist()
            lat = float('[]'.join(lati[0])) / 100
            WEi = taskfile[filenumber, 12].tolist()
            WE = ('[]'.join(WEi[0]))
            NSi = taskfile[filenumber, 10].tolist()
            NS = ('[]'.join(NSi[0]))
            time = taskfile[1, 6]
            date = str("%02d" % (int((time.split(' ')[1]).split('.')[2]),) + ':' + "%02d" % (
            int((time.split(' ')[1]).split('.')[1]),) + ':' + "%02d" % (int((time.split(' ')[1]).split('.')[0]),))
            out = MAIN + 'RGB/' + os.path.split(file)[-1][:-4] + '.jpg'

            # save jpg with exif using piexif
            exif_dict = {'0th': {},
                         '1st': {},
                         'Exif': {},
                         'GPS': {
                             1: NS,  # latituderef
                             2: ((int(lat), 1), (int(100 * (lat - int(lat))), 1),
                                 (int((100 * (lat - int(lat)) - int(100 * (lat - int(lat)))) * 6000), 100)),  # Latitude
                             3: WE,  # Longituteref
                             4: ((int(long), 1), (int(100 * (long - int(long))), 1),
                                 (int((100 * (long - int(long)) - int(100 * (long - int(long)))) * 6000), 100)),
                             # longitude
                             6: (abs(int(altitude * 100)), 100),  # Altitude
                             7: (
                             (int((time.split(' ')[2]).split(':')[0]), 1), (int((time.split(' ')[2]).split(':')[1]), 1),
                             (int((float((time.split(' ')[2]).split(':')[2])) * 100), 10)),
                             29: date},  # timestamp
                         'Interop': {},
                         'thumbnail': None}
            exif_bytes = piexif.dump(exif_dict)
            plt.imsave(out,np.dstack((band1_255, band2_255, band3_255)))
            piexif.insert(exif_bytes, out)

            """
            # save image and write EXIF information with exiftools instead
            lat=(int(lat))+(int(100*(lat-int(lat))))/60.+(100*(lat-int(lat))-int(100*(lat-int(lat))))*6000/360000.0
            long=(int(long))+(int(100*(long-int(long))))/60.+(100*(long-int(long))-int(100*(long-int(long))))*6000/360000.0
            im.save(out)
            pic = out.encode("utf-8")
            with exiftool.ExifTool() as et:
                et.execute(("-GPSLatitudeRef=" + NS).encode("utf-8"), b"-overwrite_original", pic)
                et.execute(("-GPSLongitudeRef=" + WE).encode("utf-8"),b"-overwrite_original", pic)
                et.execute(("-GPSLatitude=" + str(lat)).encode("utf-8"), b"-overwrite_original", pic)
                et.execute(("-GPSLongitude=" + str(long)).encode("utf-8"), b"-overwrite_original", pic)
                et.execute(("-GPSAltitude=" + str(altitude)).encode("utf-8"), b"-overwrite_original", pic)
                et.execute(("-GPSTimeStamp=" + str(time)).encode("utf-8"), b"-overwrite_original", pic)
                et.execute(("-GPSDateStamp=" + str(date)).encode("utf-8"), b"-overwrite_original", pic)
            """

class Rikola_RSC1( Rikola ): # old Rikola

    """
    Sensor specific details for the Rikola camera
    """
    @classmethod
    def name(cls):
        """
        Returns this sensors name
        """
        return "Rikola RSC-1"

    @classmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        return 36.5

    @classmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        return 1010

    @classmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        return 1010

    @classmethod
    def pitch(cls):
        """
        Return the pitch (mm) of the each pixel in the y-dimension (though most pixels are square).
        """
        return 0.0055

class Rikola_HSC2( Rikola ): # new Rikola

    """
    Sensor specific details for the Rikola camera
    """

    @classmethod
    def name(cls):
        """
        Returns this sensors name
        """
        return "Rikola HSC-2"

    @classmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        return 36.8

    @classmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        return 1024

    @classmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        return 1024

    @classmethod
    def pitch(cls):
        """
        Return the pitch of the each pixel in the y-dimension (though most pixels are square).
        """
        return 0.0055


"""
Functions for coregistering image and/or point cloud data.
"""

from hylite import HyImage
from hylite.project.basic import *
from hylite.project.camera import Camera
import cv2
import numpy as np
import hylite

def deepWarp(image,target):

    """
    Uses deep optical flow to warp one image to match another.

    Args:
        source (ndarray): the image to deform. Must be numpy array.
        target (ndarray): the target image to fit the source image too. Must be numpy array.

    Returns:
        A tuple containing the:

         - warped numpy array.
         - displacement map of the warp.

    """

    #convert images to greyscale uint8
    image = np.uint8(255 * (image - np.nanmin(image)) /
                         (np.nanmax(image) - np.nanmin(image)))
    target = np.uint8(255 * (target - np.nanmin(target)) /
                         (np.nanmax(target) - np.nanmin(target)))

    #calculate deep optical flow
    alg = cv2.optflow.createOptFlow_DeepFlow()
    flow = alg.calc(image,target,None)

    #distort image to fit target
    X, Y = np.meshgrid(range(image.data.shape[1]), range(image.data.shape[0]))
    dmap = np.dstack([X, Y]).astype(np.float32)
    dmap[:,:,0] += flow[:,:,0]
    dmap[:,:,1] += flow[:,:,1]

    return cv2.remap(image, dmap, None,cv2.INTER_LINEAR), dmap

def align_to_cloud_manual( cloud, cam, points, pixels, **kwds ):
    """
    Solve for camera location given a list of >4 manually chosen pixel -> point pairs.

    Args:
        cloud (hylite.HyCloud): the point cloud (HyCloud instance) to match to
        cam (hylite.project.Camera): a Camera instance containing the camera parameters (fov, etc.). The cam.pos and cam.ori properties will be ignored so can be set to anything.
        points (list): a list of keypoint ids with length > 4. The position of each keypoint should thus be given by cloud.xyz[ points[i], : ].
        pixels (list): a list of corresponding pixel coordinates (after projection), such that pixels[i] = (px,py).
        **kwds:  keywords are passed to hylite.project.pnp( ... ).

    Returns:
        A tuple containing:

         - a Camera instance containing the PnP solution.
         - err = the mean absolute error between projected keypoints (using the PnP solution) and the corresponding
                 positions given in the pixels array.
    """

    assert len(points) == len(pixels), "Error - %d pointIDs != %d corresponding pixels" % (len(points), len(pixels))
    assert len(points) >= 4, "Error - at least four pixel/point pairs are needed to solve PnP problem."

    # get world coordinate array
    points = np.array(points).astype(np.uint)
    kxyz = cloud.xyz[ points, : ]

    # get pixel coordinates array
    kxy = np.array( pixels ).astype(float) - 0.5 # must be float for OpenCV. - 0.5 transforms to pixel centers.

    # solve pnp problem
    if 'pano' in cam.proj.lower():
        kxy_pp = pano_to_persp(kxy[:,0], kxy[:, 1], cam.fov, cam.step, cam.dims)
        p_est, r_est, inl = pnp(kxyz, kxy_pp, cam.fov, cam.dims, ransac=True, **kwds)
    else:
        p_est, r_est, inl = pnp(kxyz, kxy, cam.fov, cam.dims, ransac=True, **kwds)

    # put estimate in a camera object
    est = Camera(p_est, r_est, cam.proj, cam.fov, cam.dims, cam.step)

    # calculate final correspondances
    if 'pano' in cam.proj.lower():
        pp, vis = proj_pano(kxyz[inl, :], C=p_est, a=r_est,
                            fov=cam.fov, dims=cam.dims, step=cam.step)
    else:
        pp, vis = proj_persp(kxyz[inl, :], C=p_est, a=r_est,
                             fov=cam.fov, dims=cam.dims)

    err = np.linalg.norm(pp[:, 0:2] - kxy[inl, :], axis=1)
    err = np.mean(err)

    return est, err


def refine_alignment(image, cloud, cam, bands=hylite.RGB, s=2,
                     maxdist=25, histeq=True, method='sift',
                     recurse=1, feat_args={}, match_args={}, vb=True, **kwds):
    if 'pano' in cam.proj.lower():
        assert not cam.step is None, "Error - angluar step ('step') must be defined for panoramic cameras."

    # generate rendered view to get matches from
    if vb: print("Projecting scene... ", end='')
    render = cloud.render(cam, fill_holes=True, s=s, bands='rgb')
    xyz = cloud.render(cam, fill_holes=True, s=s, bands='xyz')
    if vb: print("Done.")

    # apply masking
    image = image.export_bands(bands)
    image.set_as_nan(0)
    render.set_as_nan(0)
    mask = np.logical_or(np.isnan(render.data).all(axis=-1), np.isnan(image.data).all(axis=-1))
    mask = cv2.erode(mask.astype(np.uint8), np.ones((int(maxdist * 2) + 1, int(maxdist * 2) + 1), np.uint8))
    mask = mask == 1
    image.data[mask] = np.nan
    render.data[mask] = np.nan

    if histeq:
        from hylite.correct import hist_eq
        render.data = hist_eq(render.data, image.data)

    # extract matches
    if vb: print("Gathering matches..", end='')
    k1, d1 = image.get_keypoints((0, 1, 2), method=method, **feat_args)
    k2, d2 = render.get_keypoints((0, 1, 2), method=method, **feat_args)
    src, dst = hylite.HyImage.match_keypoints(k1, k2, d1, d2, method=method, dist=match_args.pop("dist", 0.9),
                                              **match_args)

    # filter matches by distance
    r = np.linalg.norm(src - dst, axis=-1)
    mask = r < maxdist
    src = src[mask][:, ::-1]  # flip from (y,x) to (x,y) notation here
    dst = dst[mask][:, ::-1]
    if vb: print("Found %d matches." % len(src))
    assert len(src) != 0, "Error - no matches found."

    # plot matches
    if vb:
        fig, ax = image.quick_plot((0, 1, 2), vmin=0, vmax=90)
        ax.scatter(src[:, 0], src[:, 1], color='cyan', marker='+')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlim(0, image.xdim())
        ax.set_ylim(image.ydim(), 0)
        fig.show()

    # get 3D location of tie points
    k3d = []
    for px, py in dst:
        p = xyz.data[int(px), int(py), :]  # extract 3d position
        k3d.append(p)
    k3d = np.array(k3d)

    # occasionaly keypoints have undefined positions... remove these ones.
    mask = np.isfinite(np.sum(k3d, axis=1))
    assert sum(mask) > 3, "Error - at least 4 keypoints with finite positions are needed...."
    k3d = k3d[mask, :]
    src = src[mask, :]

    if 'pano' in cam.proj.lower():
        pp = pano_to_persp(src[:, 0], src[:, 1], cam.fov, cam.step, image.data.shape)
        p_est, r_est, inl = pnp(k3d, pp, cam.fov, image.data.shape, **kwds)
    else:
        p_est, r_est, inl = pnp(k3d, src, cam.fov, image.data.shape, **kwds)

    # put estimate in a camera object
    est = Camera(p_est, r_est, cam.proj, cam.fov, image.data.shape, cam.step)

    if recurse > 1:
        return refine_alignment(image, cloud, est, bands=bands, s=s,
                                maxdist=maxdist, histeq=histeq, method=method,
                                recurse=recurse - 1, feat_args=feat_args, match_args=match_args, vb=vb, **kwds)

    # plot updated view
    if vb:
        print("Projecting updated preview...")
        render = cloud.render(est, fill_holes=True, s=s, bands='rgb')
        render.set_as_nan(0)
        fig, ax = render.quick_plot((0, 1, 2), vmin=0, vmax=90)
        ax.scatter(dst[:, 0], dst[:, 1], color='r', marker='+')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlim(0, image.xdim())
        ax.set_ylim(image.ydim(), 0)
        fig.show()

    return est, src, k3d

def align_to_cloud(image, cloud, cam, bands=hylite.RGB,
                   method='sift', recurse=2, s=2, sf=3.0, cfac=0.0, bfac=0.0,
                   vb=True, gf=True, **kwds):
    """
    Aligns a hyperspectral image to a point cloud based on an approximate camera position and
    view direction. This inital position must be reasonable to ensure good tie-points can be
    created and used to solve for the camera position.

    Args:
        image (hylite.HyImage): a HyImage object to match.
        cloud (hylite.HyCloud): a georeferenced HyCloud to match to.
        cam (hylite.project.Camera): a Camera object containing camera data (fov, step, dims) and initial position/orientation estimate.
        bands (tuple): a tuple containing the 3 hyperspectral bands to match agains the RGB bands of the point cloud. Default is
                       io.HyImage.RGB.
        method (str): the matching method to use. Can be 'sift' (default) or 'orb'.
        recurse (int): The number of rounds of matching/alignment to perform. Default is 2.
        s (int): the point size to use for rendering. Default is 2. Must be integer.
        sf (int): increase resolution of rendered images to enhance point matching. Default is 3.
        cfac (float): contrast adjustment to apply to hyperspectral bands before matching. Default is 0.0.
        bfac (float): brightness adjustment to apply to hyperspectral bands before matching. Default is 0.0.
        vb (bool): True if messages should be written to the console. Default is True.
        gf (bool): True if a plot showing matching keypoints and residuals should be created. Useful for checking
            matching accuracy, troubleshooting and evaluating the accuracy of the reconstructed camera position.
            Default is True.
        **kwds: keyword arguments are passed to the opencv feature detecting and matching algorithms.  For matching,
            selectivity can be adjusted using the dist parameter:

            - dist = the similarity threshold for identifying matches. Default is 0.7.

            For keypoint detection using SIFT, available parameters are:

            - contrastThreshold: default is 0.01.
            - edgeThreshold: default is 10.
            - sigma: default is 2.0

            For keypoint detection using ORB, available parameters are:

            - nfeatures = the number of features to detect. Default is 5000.

            Remaining keywords are passed to hylite.project.pnp( ... ).

    Returns:
        A tuple containing:

         - cam_est = a camera object with optimised positions.
         - keypoints = the keypoints used to determine this, such that keypoints[i] = [px,py,x,y,z].
         - err = average distance (in pixels) between inlier points.
    """

    if 'pano' in cam.proj.lower():
        assert not cam.step is None, "Error - angluar step ('step') must be defined for panoramic cameras."

    # generate rendered view to get matches from
    if vb: print("Projecting scene... ", end='')
    ortho_cam = cam.clone()
    ortho_cam.dims = (int(image.data.shape[0] * sf), int(image.data.shape[1] * sf)) #increase resolution by scale factor
    if ortho_cam.is_panoramic():
        ortho_cam.step = cam.step / sf
    ortho = cloud.render(ortho_cam, fill_holes=True, blur=3,
                         s=s, bands='rgbxyz')
    if vb: print("Done.")

    # extract matches
    if vb: print("Gathering matches..", end='')
    k_ortho = None
    k_hyper = None
    for b in range(3):

        # extract points
        if 'sift' in method:
            sigma = kwds.pop('sigma',2.0)
            contrastThreshold = kwds.pop('contrastThreshold',0.01)
            edgeThreshold = kwds.pop('edgeThreshold', 10)
            k1, d1 = ortho.get_keypoints(band=b, method=method, mask=True,
                                         sigma=sigma*sf, contrastThreshold=contrastThreshold,edgeThreshold=edgeThreshold,
                                         cfac=cfac, bfac=bfac)
            k2, d2 = image.get_keypoints(band=bands[b], method=method, mask=True,
                                         sigma=sigma*sf, contrastThreshold=contrastThreshold,edgeThreshold=edgeThreshold,
                                         cfac=cfac, bfac=bfac)
        elif 'orb' in method:
            nfeatures = kwds.pop('nfeatures', 5000)
            k1, d1 = ortho.get_keypoints(band=b, method=method, mask=True, nfeatures=nfeatures, cfac=cfac, bfac=bfac)
            k2, d2 = image.get_keypoints(band=bands[b], method=method, mask=True, nfeatures=nfeatures, cfac=cfac, bfac=bfac)
        else:
            assert False, "Error - unknown matching method %s" % method
        # match points
        dist = kwds.pop('dist',0.7)
        k_o, k_h = HyImage.match_keypoints(k1, k2,
                                         d1, d2, method=method, dist=dist)

        if (not k_o is None) and vb: print("..%s=%d.." % ('rgb'[b], len(k_o)), end='')

        # transform opencv matches to sensible coordinates...
        if not k_o is None:
            k_o = np.array([k_o[:, 0, 1], k_o[:, 0, 0]]).T
            k_h = np.array([k_h[:, 0, 1], k_h[:, 0, 0]]).T

        if k_ortho is None:
            k_ortho = k_o
            k_hyper = k_h
        elif not k_o is None:
            k_ortho = np.vstack([k_ortho, k_o])
            k_hyper = np.vstack([k_hyper, k_h])

    if vb and (not k_ortho is None): print("(%d)." % len(k_ortho))

    # from here on we need some keypoints....
    assert not k_ortho is None, "Error - no valid matches found..."

    # get 3D location of tie points
    k3d = []
    for px, py in k_ortho:
        p = ortho.data[int(px), int(py), 3:6]  # extract 3d position
        k3d.append(p)
    k3d = np.array(k3d)

    #occasionaly keypoints have undefined positions... remove these ones.
    mask = np.isfinite( np.sum(k3d,axis=1) )
    assert sum(mask) > 3, "Error - at least 4 keypoints with finite positions are needed...."
    k3d = k3d[mask,:]
    k_hyper = k_hyper[mask,:]

    # solve pnp problem
    try:
        if 'pano' in cam.proj.lower():
            k_hyper_pp = pano_to_persp(k_hyper[:, 0], k_hyper[:, 1], cam.fov, cam.step, image.data.shape)
            p_est, r_est, inl = pnp(k3d, k_hyper_pp, cam.fov, image.data.shape, **kwds)
        else:
            p_est, r_est, inl = pnp(k3d, k_hyper, cam.fov, image.data.shape, **kwds)

    except: # pnp failed - plot keypoints before failing
        print("Error - PnP solution not found. Check sift feature matching?")
        if gf:
            fig, ax = ortho.quick_plot((0, 1, 2))

            if not k_ortho is None:
                ax.scatter(k_ortho[:, 0], k_ortho[:, 1], c='red', s=5)
            ax.set_title("Tie points in estimated camera view.")
            fig.show()

            # plot tie points
            fig, ax = image.quick_plot(bands, cfac=cfac, bfac=bfac)
            if not k_hyper is None:
                ax.scatter(k_hyper[:, 0], k_hyper[:, 1], c='red', s=5)
            ax.set_title("Tie points in hyperspectral scene.")
            fig.show()

        return None, None, np.inf

    #put estimate in a camera object
    est = Camera(p_est,r_est,cam.proj,cam.fov,image.data.shape,cam.step)

    # calculate final correspondances
    if 'pano' in cam.proj.lower():
        pp, vis = proj_pano(k3d[inl, :], C=p_est, a=r_est,
                            fov=cam.fov, dims=image.data.shape, step=cam.step)
    else:
        pp, vis = proj_persp(k3d[inl, :], C=p_est, a=r_est,
                             fov=cam.fov, dims=image.data.shape)

    err = np.linalg.norm(pp[:, 0:2] - k_hyper[inl, :], axis=1)
    err = np.mean(err)

    if vb: print("Solved PnP problem using %d inliers (residual error = %.1f px)." % (len(inl),err))

    # recurse using updated estimates?
    if recurse > 1:
        if vb: print('------------resolving using updated camera transform------------')
        return align_to_cloud(image, cloud, est, bands=bands,
                               method=method,recurse=recurse-1, s=s, sf=sf, vb=vb, gf=gf, **kwds)

    # plot residuals
    if gf:

        # render point cloud as background (slow)
        #fig, ax = cloud.quick_plot(est,
        #                           s=1, fill_holes=True, blur=True,
        #                           bands='rgb')

        # plot original image as background (fast)
        fig, ax = image.quick_plot(bands)

        #plot tie points and residuals
        ax.scatter(pp[vis, 0], pp[vis, 1], c='r', s=2)
        ax.scatter(pp[vis, 0], pp[vis, 1], c='r', s=2)
        ax.scatter(k_hyper[inl][:, 0], k_hyper[inl][:, 1], c='cyan', s=2)
        ax.set_xlim(0, image.data.shape[0])  # ensure axis ranges haven't been changed by weird points.
        ax.set_ylim(image.data.shape[1], 1)
        for i, p in enumerate(pp):
            ax.plot([pp[i][0], k_hyper[inl][i, 0]], [pp[i][1], k_hyper[inl][i, 1]], c='white', lw=1.0)
        ax.set_title("Reprojected keypoint residuals.")
        fig.show()

    return est, k3d[inl, :], err


def align(image1, image2, source_bands, dest_bands=None, method='affine', matchdist=0.6, vb=False, **kwds):
    """
    Coregister an image or numpy array to a target image using SIFT keypoints and the specified image transform.

    Args:
        image1 (hylite.HyImage): the reference image.
        image2 (hylite.HyImage): the image to transform.
        source_bands (tuple): A tuple defining the bands to use in image 1.
        dest_bands (tuple): A tuple defining the bands to use in image 2. Defaults to source_bands.
        method (str): the method to use. Options are 'affine', 'piecewise_affine' or 'polynomial'.
        matchdist (float): the SIFT matching distance threshold. Default is 0.6.
        vb (bool): create graphical output figures for debugging. Default is False.
        **kwds*: keywords are passed to HyImage.get_keypoints( ... ).

    Returns:
        A transformed image
    """
    assert isinstance(image1, hylite.HyImage) and isinstance(image2,
                                                             hylite.HyImage), "Error - images myst be HyImage instances."

    if dest_bands is None:
        dest_bands = source_bands

    # extract features
    def getFeatures(I1, I2):
        pref = []
        ptarg = []
        for b1, b2 in zip(source_bands, dest_bands):
            k1, d1 = I1.get_keypoints(b1, method='sift', mask=True)
            k2, d2 = I2.get_keypoints(b2, method='sift', mask=True)
            kimg, ktg = I2.match_keypoints(k1, k2, d1, d2, method='sift', dist=matchdist)
            pref.append(kimg)
            ptarg.append(ktg)
        return np.vstack(pref), np.vstack(ptarg)

    pref, ptarg = getFeatures(image1, image2)

    if vb:
        fig, ax = image1.quick_plot(source_bands)
        ax.set_title("Initial features")
        ax.scatter(pref[:, 0, 1], pref[:, 0, 0], s=4, color='r')
        ax.scatter(ptarg[:, 0, 1], ptarg[:, 0, 0], s=4, color='g')
        xx = np.array([[pref[i, 0, 1], ptarg[i, 0, 1]] for i in range(pref.shape[0])]).T
        yy = np.array([[pref[i, 0, 0], ptarg[i, 0, 0]] for i in range(pref.shape[0])]).T
        ax.plot(xx, yy, color='white', alpha=0.2)
        fig.show()

    # compute transform
    try:
        from skimage import transform as tf
    except:
        assert False, "Error - please install scikit image (pip install scikit-image) to use this functionality."

    if 'piecewise' in method.lower():
        tform = tf.PiecewiseAffineTransform()
    elif 'affine' in method.lower():
        tform = tf.AffineTransform()
    elif 'poly' in method.lower():
        tform = tf.PolynomialTransform()

    tform.estimate(pref[:, 0, :], ptarg[:, 0, :])

    # apply transform
    mapped = image2.copy(data=False)
    mapped.data = tf.warp(image2.data, tform, output_shape=(image1.xdim(), image1.ydim()))

    # final plot?
    if vb:
        # get transformed matches
        pref, ptarg = getFeatures(image1, mapped)  # get new matches to show residuals
        fig, ax = mapped.quick_plot(dest_bands)
        ax.set_title("Residual features")
        ax.scatter(pref[:, 0, 1], pref[:, 0, 0], s=4, color='r')
        ax.scatter(ptarg[:, 0, 1], ptarg[:, 0, 0], s=4, color='g')
        xx = np.array([[pref[i, 0, 1], ptarg[i, 0, 1]] for i in range(pref.shape[0])]).T
        yy = np.array([[pref[i, 0, 0], ptarg[i, 0, 0]] for i in range(pref.shape[0])]).T
        ax.plot(xx, yy, color='white', alpha=0.2)
        fig.show()

    return mapped


#### DEPRECTIATED
def align_images(image1, image2, warp=True, **kwds):
    """
    Coregister an image or numpy array to a target image.

    Args:
        image1 (hylite.HyImage): the image to transform
        image2 (hylite.HyImage): the reference image to fit too.
        **kwds: Optional keywords include:

             - image_bands = the band(s) in the image to use for matching. If an integer or wavelength is
                             passed then a single band will be used. A tuple contingin a minimum index or
                             wavelength can also be passed, in which case bands will be averaged. Default is
                             (0,3) (i.e. the first 3-bands).
             - target_bands = the band(s) in the target image to use, in the same format as image_bands. Default
                              is (0,3) (i.e. the first 3-bands).
             - method = the method used for image warping. Default is 'affine'.
             - features = the feature detector to use. Can be 'sift' (default) or 'orb'.
             - warp = use dense flow to warp the image to better fit the target. Default is True.
             - dist = the distance threshold used for keypoint matching. Default is 0.75.
             - rthresh = the ransac inlier threshold for ransac outlier detection. Default is 10.0. Increase for more tolerance.

    Returns:
        A numpy array containing the coregistered image data.
    """
    print("Warning: align_images is depreciated. Please use piecewise_align, polynomial_align or affine_align instead.")

    assert isinstance(image1,HyImage) and isinstance(image2,HyImage), "Error - images myst be HyImage instances."

    #get image features
    bands = kwds.get("image_bands",(0,3))
    k_image, d_image = image1.get_keypoints(bands,method=kwds.get('features', 'sift'), mask=True)

    #get target features
    bands = kwds.get("target_bands",(0,3))
    k_target, d_target = image2.get_keypoints(bands,method=kwds.get('features', 'sift'), mask=True)

    #match features
    k_image,k_target = image2.match_keypoints(k_image, k_target,
                                              d_image, d_target,
                                              method=kwds.get('features', 'sift'), dist=kwds.get('dist',0.75))

    #filter dodgy points using ransac model
    src_pts = k_image
    dst_pts = k_target

    #filter dodgy points using ransac model
    assert (src_pts is not None) and (dst_pts is not None), "Error - no valid matches found."
    H, status = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, kwds.get('rthresh', 10.0))
    dst_mask = dst_pts[:, 0, :] * status
    src_mask = src_pts[:, 0, :] * status
    dst_mask = dst_mask[dst_mask.all(1)]
    src_mask = src_mask[src_mask.all(1)]

    method = kwds.get('method', 'affine').lower() # get transform method
    if 'affine' in method: # warp with affine transform
        dst_mask = np.expand_dims(dst_mask, axis=1)
        src_mask = np.expand_dims(src_mask, axis=1)
        #M = cv2.estimateRigidTransform(src_mask, dst_mask, False) # estimate affine transform
        M = cv2.estimateAffinePartial2D(src_mask, dst_mask)[0]

        # if > 512 bands, cut in half for open-cv compatability
        if image1.band_count() > 1025:
            assert False, "Too many baaaands!!!"
        if image1.band_count() < 512: # easy
            #mapped = cv2.warpPerspective(image1.data, M, (image2.ydim(), image2.xdim()))
            mapped = cv2.warpAffine(image1.data, M, (image2.data.shape[1], image2.data.shape[0])) # apply
        else: # slightly less easy
            div = int(image1.data.shape[2] / 2) # split dataset

            # map each part
            mapped1 = cv2.warpAffine(image1.data[...,:div], M, (image2.data.shape[1], image2.data.shape[0]))
            mapped2 = cv2.warpAffine(image1.data[...,div:], M, (image2.data.shape[1], image2.data.shape[0]))
            #mapped1 = cv2.warpPerspective(image1.data[..., :div], H, (image2.ydim(), image2.xdim()))
            #mapped2 = cv2.warpPerspective(image1.data[..., div:], H, (image2.ydim(), image2.xdim()))

            # rejoin
            mapped = np.concatenate((mapped1, mapped2), axis=2)

    elif 'poly' in method: # warp with polynomial
        try:
            from skimage import transform as tf
        except:
            assert False, "Error - please install scikit image to use poly mode: pip install scikit-image"
        tform3 = tf.estimate_transform('polynomial', dst_mask, src_mask)
        mapped = tf.warp(image1.data, tform3, output_shape=(image2.xdim(), image2.ydim()))
    else:
        assert False, "Unknown transform method %s." % method

    #refine using deep flow?
    if warp:
        _, dmap = deepWarp( np.nanmean( mapped, axis=2 ), np.nanmean( image2.data, axis=2) )
        mapped = cv2.remap( mapped, dmap, None, cv2.INTER_LINEAR)
    return mapped





"""
Import and export hyperspectral data. For hyperspectral images this is mostly done using GDAL,
while for point clouds and hyperspectral libraries a variety of different methods are included.
"""
from .headers import *
from .images import *
from .clouds import *
from .libraries import *
from .pmaps import *
from .cameras import saveCameraTXT, loadCameraTXT

from hylite import HyImage, HyCloud, HyLibrary, HyCollection, HyScene, HyData
from hylite.project import PMap, Camera, Pushbroom
from hylite.analyse.mwl import MWL
from distutils.dir_util import copy_tree
import os

# check if gdal is installed
try:
    from osgeo import gdal
    usegdal = True
except ModuleNotFoundError:
    usegdal = False

def save(path, data, **kwds):
    """
    A generic function for saving HyData instances such as HyImage, HyLibrary and HyCloud. The appropriate file format
    will be chosen automatically.

    Args:
        path (str): the path to save the file too.
        data (HyData or ndarray): the data to save. This must be an instance of HyImage, HyLibrary or HyCloud.
        **kwds: Keywords can include:

             - vmin = the data value that = 0 when saving RGB images.
             - vmax = the data value that = 255 when saving RGB images. Must be > vmin.
    """

    if isinstance(data, HyImage):

        # special case - save ternary image to png or jpg or bmp
        ext = os.path.splitext(path)[1].lower()
        if 'jpg' in ext or 'bmp' in ext or 'png' in ext or 'pdf' in ext:
            if data.band_count() == 1 or data.band_count() == 3 or data.band_count == 4:
                rgb = np.transpose( data.data, (1,0,2) )
                if not ((data.is_int() and np.max(rgb) <= 255)): # handle normalisation
                    vmin = kwds.get("vmin", np.nanpercentile(rgb, 1 ) )
                    vmax = kwds.get("vmax", np.nanpercentile(rgb, 99) )
                    rgb = (rgb - vmin) / (vmax-vmin)
                    rgb = (np.clip(rgb, 0, 1) * 255).astype(np.uint8) # convert to 8 bit image
                #from matplotlib.pyplot import imsave
                # imsave( path, rgb )
                from skimage import io as skio
                skio.imsave( path, rgb ) # save the image
                return
        elif ((data.band_count() == 1) or (data.band_count() == 3) or (data.band_count() == 4)) and (data.data.dtype == np.uint8):
            # save 1, 3 and 4 band uint8 arrays as png files
            # from matplotlib.pyplot import imsave
            # imsave( os.path.splitext(path)[0]+".png", data.data)  # save the image
            #from skimage import io as skio
            #skio.imsave(os.path.splitext(path)[0]+".png", np.transpose( data.data, (1,0,2) ))  # save the image
            from PIL import Image
            if (data.band_count() == 1):
                img = Image.fromarray(data.data[..., 0].T)  # single-band PNG
            else:
                img = Image.fromarray(np.transpose(data.data, (1, 0, 2)))  # ternary PNG

            img.save(os.path.splitext(path)[0]+".png", "PNG")
            save( os.path.splitext(path)[0] + ".hdr", data.header ) # save header
            return
        else: # save hyperspectral image
            if usegdal:
                from osgeo import gdal  # is gdal installed?
                save_func = saveWithGDAL
            else:  # no gdal, use SPy
                save_func = saveWithSPy
            if 'lib' in ext: # special case - we are actually saving a HyLibrary (as an image)
                ext = 'lib'
            else:
                ext = 'dat'
    elif isinstance(data, HyHeader):
        save_func = saveHeader
        ext = 'hdr'
    elif isinstance(data, HyCloud):
        save_func = saveCloudPLY
        ext = 'ply'
    elif isinstance(data, HyLibrary):
        save_func = saveLibraryLIB
        ext = 'lib'
    elif isinstance(data, PMap ):
        save_func = savePMap
        ext = 'npz'
    elif isinstance(data, Camera ):
        save_func = saveCameraTXT
        ext = 'cam'
    elif isinstance(data, Pushbroom):
        save_func = saveCameraTXT
        ext = 'brm'
    elif isinstance(data, HyCollection):
        save_func = _saveCollection
        ext = data.ext[1:]
        outdir = os.path.join(data.root, os.path.splitext(data.name)[0])
        os.makedirs( os.path.splitext(path)[0] +"."+ ext, exist_ok=True ) # make output directory (even if empty)
        if os.path.splitext(path)[0] != outdir:
            if os.path.exists( outdir+"."+ext): # if it exists...
                #if sys.version_info[1] >= 8: # python 3.8 or greater
                #    shutil.copytree( outdir+"."+ext, os.path.splitext(path)[0]+"."+ext, dirs_exist_ok=True)
                #else:
                #    shutil.copytree( outdir+"."+ext, os.path.splitext(path)[0]+"."+ext ) # will fail if directory already exists unfortunately.
                copy_tree(outdir+"."+ext, os.path.splitext(path)[0]+"."+ext)


    elif isinstance(data, np.ndarray) or isinstance(data, list):
        save_func = np.save
        ext = 'npy'
    else:
        assert False, "Error - data type %s is unsupported by hylite.io.save." % type(data)

    # check path file extension
    if 'hdr' in os.path.splitext(path)[1]: # auto strip .hdr extensions if provided
        path = os.path.splitext(path)[0]
    if ext not in os.path.splitext(path)[1]: # add type-specific extension if needed
        path += '.%s'%ext

    # save!
    os.makedirs( os.path.dirname(path), exist_ok=True)  # make output directory
    save_func( path, data )

def load(path):
    """
    A generic function for loading hyperspectral images, point clouds and libraries. The appropriate load function
    will be chosen based on the file extension.

    Args:
        path (str): the path of the file to load.

    Returns:
        The loaded data.
    """

    assert os.path.exists( path ), "Error: file %s does not exist." % path

    # load file formats with no associated header
    if 'npz' in os.path.splitext( path )[1].lower():
        return loadPMap(path)
    elif 'npy' in os.path.splitext( path )[1].lower():
        return np.load( path ) # load numpy

    # file (should/could) have header - look for it
    header, data = matchHeader( path )
    assert os.path.exists(str(data)), "Error - data file %s does not exist." % path
    ext = os.path.splitext(data)[1].lower()
    if ext == '':
        assert os.path.isfile(data), "Error - %s is a directory not a file." % data

    # load other file types
    if 'ply' in ext: # point or hypercloud
        out = loadCloudPLY(path) # load dataset
    elif 'las' in ext: # point or hypercloud
        out =  loadCloudLAS(path)
    elif 'csv' in ext: # (flat) spectral library
        out = loadLibraryCSV(path)
    elif 'txt' in ext: # (flat) spectral library
        out = loadLibraryTXT(path)
    elif 'sed' in ext: # (flat) spectral library
        out = loadLibrarySED(path)
    elif 'tsg' in ext: # (flat) spectral library
        out = loadLibraryTSG(path)
    elif 'hyc' in ext or 'hys' in ext or 'mwl' in ext: # load hylite collection, hyscene or mwl map
        out = _loadCollection(path)
    elif 'cam' in ext or 'brm' in ext: # load pushbroom and normal cameras
        out = loadCameraTXT(path)
    else: # image
        # load conventional images with PIL
        if 'png' in ext or 'jpg' in ext or 'bmp' in ext:
            # load image with matplotlib
            #from matplotlib.pyplot import imread
            #im = imread(path)
            from skimage import io as skio
            im = skio.imread(data)
            if len(im.shape) == 2:
                im = im[:,:,None] # add last dimension if greyscale image is loaded
            out = HyImage(np.transpose(im, (1, 0, 2)))
            if header is not None:
                out.header = loadHeader(header)
        else:
            if usegdal:
                from osgeo import gdal # is gdal installed?
                out = loadWithGDAL(path)
            else: # no gdal, use SPy
                out = loadWithSPy(path)

        # special case - loading spectral library; convert image to HyData
        if 'lib' in ext:
            out = HyLibrary(out.data, header=out.header)
    return out  # return dataset

##############################################
## save and load data collections
##############################################
# save collection
def _saveCollection(path, collection):
    # generate file paths
    dirmap = collection.get_file_dictionary(root=os.path.dirname(path),
                                            name=os.path.splitext(os.path.basename(path))[0])
    # print(dirmap)
    for p, o in dirmap.items():
        os.makedirs(os.path.dirname(p), exist_ok=True) # make output directory if needed
        save(p, o)  # save each path and item [ n.b. this includes the header file! :-) ]

def _loadCollection(path):
    # load header and find directory path
    header, directory = matchHeader(path)

    # parse name and root
    root = os.path.dirname(directory)
    name = os.path.basename(os.path.splitext(directory)[0])

    if 'hyc' in os.path.splitext(directory)[1]:
        C = HyCollection(name, root, header=loadHeader(header))
    elif 'hys' in os.path.splitext(directory)[1]:
        C = HyScene(name, root, header=loadHeader(header))
    elif 'mwl' in os.path.splitext(directory)[1]:
        C = MWL(name, root, header=loadHeader(header))
    else:
        # print(header, directory )
        assert False, "Error - %s is an invalid collection." % directory
    return C


"""
Projection maps store lookup tables (python dictionaries) that link PointIDs in a point cloud with pixelIDs in an image.
They store many : many relationships and can store arbitrarily complicated projections ( perspective, panoramic,
pushbroom etc.). PMaps only store the mapping function; see HyScene for functionality that pushes data between different
data types. PMaps can be saved using io.save( ... ) to avoid expensive computation multiple times in a processing
workflow.
"""

import os
import numpy as np
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix
import hylite
from tempfile import mkdtemp
import shutil
from tqdm import tqdm

class PMap(object):
    """
    A class for storing lookup tables that map between 3-D point clouds and 2-D images.
    """

    def __init__(self, xdim, ydim, npoints, cloud=None, image=None):
        """
        Create a new (empty) PMap object.

        Args:
             xdim: the width of the associated image in pixels. Used for ravelling indices.
             ydim: the height of the associated image in pixels. Used for ravelling indices.
             points: the number of points that are in this map. Used for ravelling indices.
             cloud: a link to the source cloud, default is None.
             image: a link to the source image, default is None.
        """
        self.xdim = xdim  # width of the associated image
        self.ydim = ydim  # height of the associated image
        self.npoints = npoints
        self.cloud = cloud # store ref to cloud if provided
        self.image = image # store ref to image if provided
        self.points = [] # points with non-zero references in this matrix
        self.pixels = [] # pixels with non-zero references in this matrix
        self.data = coo_matrix( (npoints, xdim*ydim), dtype=np.float32 ) # initialise mapping matrix

    def _ridx(self, pixel):
        """
        Ravel a pixel index tuple to an integer. If an integer is passed to this function
        it will be returned directly.
        """
        if isinstance(pixel, tuple):
            assert len(pixel) == 2, "Error - (x,y) tuples must have length 2."
            pixel = np.ravel_multi_index(pixel, dims=(self.xdim, self.ydim), order='F')
        assert np.issubdtype(type(pixel), np.integer), "Error - non-integer pixel ID (%s = %s)?" % (
        pixel, type(pixel))  # always wear protection
        return pixel

    def _uidx(self, pixel):
        """
        Unravel an integer pixel index to a (x,y) tuple. If a tuple is passed to this function
        it will be returned directly.
        """
        if np.issubdtype(type(pixel), np.integer):
            pixel = np.unravel_index(pixel, (self.xdim, self.ydim), order='F')
        elif isinstance(pixel, tuple):
            pass
        else:
            assert np.issubdtype(type(pixel), np.integer), "Error - non-integer pixel ID (%s = %s)?" % (
            pixel, type(pixel))
        return pixel

    def coo(self):
        """
        Convert this pmap's internal sparse matrix to coo format.
        """
        if not isinstance(self.data, coo_matrix):
            self.data = self.data.tocoo()
    def csc(self):
        """
        Convert this pmap's internal sparse matrix to compressed column format.
        """
        if not isinstance(self.data, csc_matrix):
            self.data = self.data.tocsc()
    def csr(self):
        """
        Convert this pmap's internal sparse matrix to compressed row format.
        """
        if not isinstance(self.data, csr_matrix):
            self.data = self.data.tocsr()

    def get_flat(self):
        """
        Return three flat arrays that contain all the links in this projection map.

        Returns:
            A tuple containing:

            - points = an (n,) list of point ids.
            - pixels = an (n,) list of pixel ids corresponding to the points above.
            - z = an (n,) list of distances between each point and corresponding pixel.
        """
        self.coo()
        return self.data.row, self.data.col, 1/self.data.data # return

    def set_flat(self, points, pixels, z ):
        """
        Adds links to this pmap from flattened arrays as returned by get_flat( ... ).

        Args:
            points: an (n,) list of point ids.
            pixels: an (n,) list of pixel ids corresponding to the points above.
            z: an (n,) list of distances between each point and corresponding pixel.
        """
        # easy!
        self.data = coo_matrix( (1/z, (points,pixels)), shape=(self.npoints, self.xdim*self.ydim), dtype=np.float32 )

        # also store unique points and pixels
        self.points = np.unique( points )
        self.pixels = np.unique( pixels )

    def set_ppc(self, pp, vis):
        """
        Adds links to the pmap based on a projected point coordinates array and a visibility list, as returned by
        e.g. proj_persp and proj_pano.

        Args:
            pp: projected point coordinates as returned by e.g. proj_pano.
            vis: point visibilities, as returned by e.g. proj_pano.
        """
        # convert to indices
        pid = np.argwhere(vis)[:, 0]
        pp = pp[vis]

        # convert pixel indices to flat indices
        pix = np.ravel_multi_index(pp[:,[0,1]].astype(int).T, dims=(self.xdim, self.ydim), order='F')

        # set data values
        self.set_flat( pid, pix, pp[:,2] )

    def size(self):
        """
        How many relationships are stored in this?
        """
        return len( self.data.data )

    def point_count(self):
        """
        How many points are included in this mapping (i.e. how many points are mapped to pixels?)
        """
        return len(self.points)

    def pixel_count(self):
        """
        How many pixels are included in this mapping?
        """
        return len(self.pixels)

    def get_depth(self, pixel, point):
        """
        Get the distance between a pixel and point pair. Returns
        None if there is no mapping between the pixel and the point.
        """
        pixel = self._ridx(pixel)
        if isinstance(self.data, coo_matrix): # need to change from coo coordiantes
            self.csc()
        return 1 / self.data[ point, pixel ] # n.b. note that matrix entries are 1 / z

    def get_point_index(self, pixel):
        """
        Get the index of the closest point in the specified pixel

        Args:
            pixel: the index of the pixel (integer or (x,y) tuple).
        Returns:
            A tuple containing:

             - point = the point index or None if no points are in the specified pixel.
             - depth = the distance to this point.
        """
        self.csc() # convert to column format
        C = self.data[:, self._ridx(pixel)] # get column
        return C.nonzero()[0][ np.argmax( C.data ) ], 1 / np.max( C.data ) # return closest point
                                                                       # n.b. note that matrix entries are 1 / z

    def get_point_indices(self, pixel):
        """
        Get the indices and depths of all points in the specified pixel.

        Args:
            pixel: the index of the pixel (integer or (x,y) tuple).
        Returns:
            A tuple containing:
                 - points = a list of point indices, or [ ] if no points are present.
                 - depths = a list of associated depths, or [ ] if no points are present.
        """
        self.csc() # convert to column format
        C = self.data[:, self._ridx(pixel)] # get column
        return C.nonzero()[0], 1 / C.data # return indices and depths

    def get_pixel_index(self, point):
        """
        Get the index of the closest pixel to the specified point.

        Args:
            point: the point index
        Returns:
            A tuple containing:

             - (px, py) = the pixel coordinates, or None if no mapping exists
             - depth = the distance to this pixel.
        """
        self.csr() # convert to row format
        R = self.data[point, :] # get row
        return self._uidx( R.nonzero()[1][ np.argmax( R.data ) ] ), 1 / np.max( R.data ) # closest pixel and depth

    def get_pixel_indices(self, point):
        """
        Get a list of pixel coordinates associated with the specified point.

        Args:
            point: the point index

        Returns:
            A tuple containing:

                 - pixels = a list of (n,2) containing pixel coordinates.
                 - depths = a list of (n,) containing associated distances.
        """
        self.csr() # convert to row format
        R = self.data[point,:] # get row
        return np.array( np.unravel_index(R.nonzero()[1], (self.xdim, self.ydim), order='F')).T, R.data

    def get_pixel_depths(self):
        """
        Return a (xdim,ydim) array containing the depth to the closest point in each pixel. Pixels with no points
        will be given 0 values.
        """
        out = (1 / np.max(self.data, axis=0).toarray()).astype(np.float32)
        out[np.logical_not(np.isfinite(out))] = 0
        return out.reshape(self.xdim,self.ydim,order='F')

    def get_point_depths(self):
        """
        Return a (npoints,) array containing the depth to the closest pixel from each point. Points with no
        pixels will be given 0 values.
        """
        out = (1 / np.max(self.data, axis=1).toarray()).astype(np.float32)
        out[np.logical_not(np.isfinite(out))] = 0
        return out

    def points_per_pixel(self):
        """
        Calculate how many points are in each pixel.

        Returns:
            a HyImage instance containing point counts per pixel.
        """
        self.csr() # use row-compressed form
        W = (self.data > 0).astype(np.float32)  # convert to binary adjacency matrix
        npnt = np.array(W.sum(axis=0)).ravel()  # get number of points per pixel
        return hylite.HyImage( npnt.reshape( (self.xdim, self.ydim, 1 ), order='F' ) )

    def pixels_per_point(self):
        """
        Calculates how many pixels project to each point.

        Returns:
            a copy of self.cloud, but with a scalar field containing pixel counts per point. If self.cloud is not defined
           then a numpy array of point counts will be returned.
        """
        self.csc() # use column compressed form
        W = (self.data > 0).astype(np.float32)  # convert to binary adjacency matrix
        npix = np.array(W.sum(axis=1)).ravel()  # get number of points per pixel

        if self.cloud is not None: # return a cloud
            out = self.cloud.copy( data = False )
            out.data = npix[:,None]
            return out
        else:
            return npix # return a numpy array

    def intersect(self, map2):
        """
        Get point indices that exist (are visible in) this scene and another.

        Args:
            map2: a PMap instance that references the same cloud but with a different image/viewpoint.
        Returns:
            a list of point indices that are visible in both scenes.
        """
        S1 = set( self.points )
        S2 = set( map2.points )
        return list( S1 & S2 )

    def union(self, maps):
        """
        Returns points that are included in one or more of the passed PMaps.

        Args:
            maps: a list of pmap instances to compare with (or just one).
        Returns:
            a list of point indices that are visible in either or both scenes.
        """
        if not isinstance(maps, list):
            maps = [maps]
        S_n = [set(s.points) for s in maps]
        S1 = set(self.points)
        return S1.union(*S_n)

    def intersect_pixels(self, map2):
        """
        Identifies matching pixels between two scenes.

        Args:
            map2: the scene to match against this one

        Returns:
            A tuple containing:

             - px1 = a numpy array of (x,y) pixel coordinates in this scene.
             - px2 = a numpy array of corresponding (x,y) pixel coordinates in scene 2.
        """
        px1 = []
        px2 = []
        overlap = self.intersect(map2)  # get points visible in both
        for idx in overlap:
            for _px1 in self.get_pixel_indices(idx)[0]:
                for _px2 in map2.get_pixel_indices(idx)[0]:
                    px1.append(_px1)
                    px2.append(_px2)
        return np.array(px1), np.array(px2)

    def remove_nan_pixels(self, image=None):
        """
        Removes mappings to nan pixels from linkage matrix.

        Args:
            image: the image containing nan pixels that should be removed. Default is self.image.
        """
        self.csc()  # change to column format

        if image is None:
            image = self.image

        # build list of nan columns
        isnan = np.logical_not(np.isfinite(image.data).all(axis=-1)).ravel(order='F')
        f = np.ones(isnan.shape[0])
        f[isnan] = 0

        # zero elements in these
        self.data = self.data.multiply(f)

        # remove zero elements
        self.data.eliminate_zeros()

    def filter_footprint(self, thresh=50):
        """
        Filter projections in a PMap instance and remove pixels that have a
        on-ground footprint above the specified threshold. This operation is
        applied in-place to conserve memory.

        Args:
            thresh: the maximum allowable pixel footprint (in points). Pixels containing > than
                    this number of points will be removed from the projection map.
        """

        # calculate footprint
        W = (self.data > 0).astype(np.float32)  # convert to binary adjacency matrix
        n = np.array(W.sum(axis=0)).ravel()  # get number of points per pixel

        # if isinstance(thresh, int): # calculate threshold as percentile if need be
        #    thresh = np.percentile( n[ n > 0], thresh )

        # convert to coo format
        self.coo()

        # rebuild mapping matrix
        mask = n[self.data.col] < thresh  # valid points
        self.data = coo_matrix((self.data.data[mask], (self.data.row[mask], self.data.col[mask])),
                       shape=(self.npoints, self.xdim * self.ydim), dtype=np.float32)

    def filter_occlusions(self, occ_tol=5.):
        """
        Filter projections in a PMap instance and remove points that are likely to be
        occluded. This operation is applied in-place to conserve memory.

        Args:
            occ_tol: the tolerance of the occlusion culling. Points within this distance of the
                     closest point in each pixel will be retained.
        """

        zz = np.max(self.data, axis=0).power(-1)  ## calculate closest point in each pixel
        zz.data += occ_tol
        zz = zz.tocsc()
        self.coo()

        # rebuild mapping matrix
        mask = self.data.data > (1 / zz[0, self.data.col].toarray()[0, :])  ## which points to include
        self.data = coo_matrix((self.data.data[mask], (self.data.row[mask], self.data.col[mask])),
                       shape=(self.npoints, self.xdim * self.ydim), dtype=np.float32)


def _gather_bands(data, bands):
    """
    Utility function used by push_to_image( ... ) and push_to_cloud( ... ) to slice data from a HyData instance.

    Returns:
        A tuple containing:

             - data = a data array containing the requested bands (hopefully).
             - wav = the wavelengths of the extracted bands (or -1 for non-spectral attributes).
             - names = the names of the extracted bands.
    """

    # extract wavelength and band name info
    dat = []
    wav = []
    nam = []

    # loop through bands tuple/list and extract data indices/slices
    for e in bands:
        # extract from point cloud based on string
        if isinstance(e, str):
            for c in e.lower():
                if c == 'r':
                    assert data.has_rgb(), "Error - RGB information not found."
                    dat.append(data.rgb[..., 0])
                    nam.append('r')
                    wav.append(hylite.RGB[0])
                elif c == 'g':
                    assert data.has_rgb(), "Error - RGB information not found."
                    dat.append(data.rgb[..., 1])
                    nam.append('g')
                    wav.append(hylite.RGB[1])
                elif c == 'b':
                    assert data.has_rgb(), "Error - RGB information not found."
                    dat.append(data.rgb[..., 2])
                    nam.append('b')
                    wav.append(hylite.RGB[2])
                elif c == 'x':
                    dat.append(data.xyz[..., 0])
                    nam.append('x')
                    wav.append(-1)
                elif c == 'y':
                    dat.append(data.xyz[..., 1])
                    nam.append('y')
                    wav.append(-1)
                elif c == 'z':
                    dat.append(data.xyz[..., 2])
                    nam.append('z')
                    wav.append(-1)
                elif c == 'k':
                    assert data.has_normals(), "Error - normals not found."
                    dat.append(data.normals[..., 0])
                    nam.append('k')
                    wav.append(-1)
                elif c == 'l':
                    assert data.has_normals(), "Error - normals not found."
                    dat.append(data.normals[..., 1])
                    nam.append('l')
                    wav.append(-1)
                elif c == 'm':
                    assert data.has_normals(), "Error - normals not found."
                    dat.append(data.normals[..., 2])
                    nam.append('m')
                    wav.append(-1)
        # extract slice
        elif isinstance(e, tuple):
            assert len(e) == 2, "Error - band slices must be tuples of length two."
            idx0 = data.get_band_index(e[0])
            idx1 = data.get_band_index(e[1])
            dat += [data.data[..., b] for b in range(idx0, idx1)]
            if data.has_band_names():
                nam += [data.get_band_names()[b] for b in range(idx0, idx1)]
            else:
                nam += [str(b) for b in range(idx0, idx1)]
            if data.has_wavelengths():
                wav += [data.get_wavelengths()[b] for b in range(idx0, idx1)]
            else:
                wav += [float(b) for b in range(idx0, idx1)]
        # extract band based on index or wavelength
        elif isinstance(e, float) or isinstance(e, int):
            b = data.get_band_index(e)
            dat.append(data[..., b])
            if data.has_band_names():
                nam.append(data.get_band_names()[b])
            else:
                nam.append(str(b))
            if data.has_wavelengths():
                wav.append(data.get_wavelengths()[b])
            else:
                wav.append(float(b))
        else:
            assert False, "Unrecognised band descriptor %s" % b

    if data.is_image():
        dat = np.dstack(dat)  # stack
    else:
        dat = np.vstack(dat).T
    return dat, wav, nam

def push_to_cloud(pmap, bands=(0, -1), method='best', image=None, cloud=None ):
    """
    Push the specified bands from an image onto a hypercloud using a (precalculated) PMap instance.

    Args:
        pmap: a pmap instance. the pmap.image and pmap.cloud references must also be defined.
        bands: List defining the bands to include in the output dataset. Elements should be one of:

            - numeric = index (int), wavelength (float) of an image band
            - tuple of length 2: start and end bands (float or integer) to export.
            - iterable of length > 2: list of bands (float or integer) to export.
        method: The method used to condense data from multiple pixels onto each point. Options are:

            - 'closest': use the closest pixel to each point.
            - 'distance': average with inverse distance weighting.
            - 'count' : average weighted inverse to the number of points in each pixel.
            - 'best' : use the pixel that is mapped to the fewest points (only). Default.
            - 'average' : average with all pixels weighted equally.

        image: the image to project (if different to pmap.image). Must have matching dimensions. Default is pmap.image.
        cloud: the cloud to project (if different to pmap.cloud). Must have matching dimensions. Default is pmap.cloud.

    Returns:
        A HyCloud instance containing the back-projected data.
    """

    if image is None:
        image = pmap.image
    if cloud is None:
        cloud = pmap.cloud

    # get image array of data to copy across
    data = image.export_bands(bands)

    # flatten it
    X = data.data.reshape(-1, data.data.shape[-1], order='F')  # n.b. we don't used data.X() to ensure 'C' order

    # convert pmap to csc format
    pmap.csc()  # thunk; would csr format be faster??
    pmap.remove_nan_pixels(image=image) # drop nan pixels

    # build weights matrix
    if 'closest' in method.lower():
        # get closest pixels
        closest = np.argmax(pmap.data, axis=1)  # find closest pixel to each point (largest 1/z along cols)

        # assemble sparse matrix with closest pixels scored as 1
        rows = np.argwhere(closest > 0)[:, 0]
        cols = np.array(closest[rows])[:, 0]
        vals = np.ones(rows.shape)
        W = csc_matrix((vals, (rows, cols)), pmap.data.shape, dtype=np.float32)

        # sum of weights is easy
        n = np.ones(pmap.npoints, dtype=np.float32)
    elif 'average' in method.lower():
        W = (pmap.data > 0).astype(np.float32)  # weights matrix [ full of ones ]
        n = np.array(W.sum(axis=1))[:, 0]  # sum of weights (for normalising later)
    elif 'count' in method.lower() or 'best' in method.lower():
        W = (pmap.data > 0).astype(np.float32)  # convert to binary adjacency matrix
        npoints = np.array(W.sum(axis=0))[0, :]  # get number of points per pixel
        W = W.multiply(1 / npoints)  # fill non-zero values with 1 / points in relevant pixel
        if 'best' in method.lower():  # filter W so we keep only the best points
            best = np.argmax(W, axis=1)  # assemble sparse matrix with best pixels scored as 1
            rows = np.argwhere(best > 0)[:, 0]
            cols = np.array(best[rows])[:, 0]
            vals = np.ones(rows.shape)
            W = csc_matrix((vals, (rows, cols)), pmap.data.shape, dtype=np.float32)

        n = np.array(W.sum(axis=1))[:, 0]  # sum of weights (for normalising later)
    elif 'distance' in method.lower():
        W = pmap.data  # easy!
        n = np.array(W.sum(axis=1))[:, 0]  # sum of weights (for normalising later)
    else:
        assert False, "Error - %s is an invalid method." % method

    # calculate output
    #V = W.dot(X) / n[:, None]
    V = W@X / n[:, None]

    # build output cloud
    out = cloud.copy(data=False)
    out.data = V
    if data.has_wavelengths():
        out.set_wavelengths(data.get_wavelengths())

    return out


def push_to_image(pmap, bands='xyz', method='closest', image=None, cloud=None):
    """
    Project the specified data from a point cloud onto an image using a (precalculated) PMap instance. If multiple points map
    to a single pixel then the results are averaged.

    Args:
        pmap: a pmap instance. the pmap.image and pmap.cloud references must also be defined.
        bands: List defining the bands to include in the output dataset. Elements should be one of:

            - numeric = index (int), wavelength (float) of an image band
            - bands = a list of image band indices (int) or wavelengths (float). Inherent properties of point clouds
               can also be expected by passing any combination of the following:
                - 'rgb' = red, green and blue per-point colour values
                - 'klm' = point normals
                - 'xyz' = point coordinates
            - iterable of length > 2: list of bands (float or integer) to export.

        method: The method used to condense data from multiple points onto each pixel. Options are:

            - 'closest': use the closest point to each pixel (default is this is fastest).
            - 'average' : average with all pixels weighted equally. Slow.

        image: the image to project (if different to pmap.image). Must have matching dimensions. Default is pmap.image.
        cloud: the cloud to project (if different to pmap.cloud). Must have matching dimensions. Default is pmap.cloud.
    Returns:
        A HyImage instance containing the projected data.
    """

    if image is None:
        image = pmap.image
    if cloud is None:
        cloud = pmap.cloud

    # special case: individual band; wrap in list
    if isinstance(bands, int) or isinstance(bands, float) or isinstance(bands, str):
        bands = [bands]

        # special case: tuple of two bands; treat as slice
        if isinstance(bands, tuple) and len(bands) == 2:
            bands = [bands]

    # gather data to project
    dat, wav, nam = _gather_bands(cloud, bands)  # extract point normals, positions and sky view

    # convert pmap to csr format
    # pmap.csr()

    # build weights matrix
    if 'closest' in method.lower():
        closest = np.array(np.argmax(pmap.data, axis=0))[0,
                  :]  # find closest point to each pixel (largest 1/z along rows)

        # assemble sparse matrix with closest points scored as 1
        cols = np.argwhere(closest > 0)[:, 0]
        rows = np.array(closest[cols])
        vals = np.ones(rows.shape)
        W = csc_matrix((vals, (rows, cols)), pmap.data.shape, dtype=np.float32)
        #V = W.T.dot(dat)  # project closest poits
        V = W.T@dat  # project closest poits
    elif 'average' in method.lower():
        W = (pmap.data > 0).astype(np.float32)  # weights matrix [ full of ones ]
        n = np.array(W.sum(axis=0))[0, :]  # sum of weights
        #V = W.T.dot(dat) / n[:, None]  # calculate average
        V = W.T@dat / n[:, None]  # calculate average
    else:
        assert False, "Error - %s is an invalid method for cloud_to_image." % method

    # build output image
    out = image.copy(data=False)
    out.data = np.reshape(V, (image.xdim(), image.ydim(), -1), order='F')
    out.data[out.data == 0] = np.nan  # replace zeros with nans
    out.set_wavelengths(wav)
    out.set_band_names(nam)

    return out

# functions for blending scenes together

def push_geomattr(scene, method='best'):
    """
    Return a HyCloud instance containing the geometric attributes of a given projection. This will have
    two or three bands as follows:

     - distance from the sensor to each point
     - effective GSD (number of points per source pixel)
     - obliquity (angle in degrees between surface normal and view vector), if the point cloud has normal vectors.

    These are useful for QAQC of projected results or doing weighted blending.

    Args:
        scene: the HyScene instance to compute geometric attributes for.
        method: the projection method used for handling duplicate pixels. See scene.push_to_cloud for details.
    Returns:
        a HyCloud instance containing the three geometric attributes.
    """
    # get projection map and relevant attributes (in image space)
    pmap = scene.pmap
    attr = [scene.depth,  # point depth
            pmap.points_per_pixel().data[..., 0]]  # gsd
    bands = ['depth', 'gsd']
    if hasattr(scene, "normals"):
        # dot product of view product and surface normal vector
        attr.append(np.rad2deg(np.arccos(
            -(scene.view * scene.normals).sum(-1))))
        bands.append('obliquity')

    # replace zeros with nans
    for a in attr:
        a[a == 0] = np.nan
    image = hylite.HyImage(np.dstack(attr))
    attr = scene.push_to_cloud((0, -1), method=method, image=image)
    attr.set_band_names(bands)
    return attr


def get_blend_weights(scenes, method, ascloud=True):
    """
    Compute weights for doing scene blending based on geometric attributes computed using
    push_geomattr( ... ).

    Args:
        scenes: a list of scenes to compute blend weights for.
        method: the weighting method. Options are:

             - 'equal' = all values are weighted equally.
             - 'distance' = closer points are given higher weight.
             - 'gsd' = lower ground-sampling values are given higher weight.
             - 'obliquity' = less oblique points are given higher weight.

        ascloud: True if weights should be returned as a HyCloud instance. Otherwise a numpy array is returned.
    Returns:
        a numpy array containing the normalised blending weight for each (point,scene).
    """

    weights = np.ones((scenes[0].cloud.point_count(), len(scenes)))
    if 'equal' not in method.lower():
        for i, s in enumerate(scenes):
            # compute geometric features
            geom = push_geomattr(s)
            if 'dist' in method.lower():
                weights[:, i] = geom.data[:, 0]
            elif 'gsd' in method.lower():
                weights[:, i] = geom.data[:, 1]
            elif 'obl' in method.lower():
                weights[:, i] = geom.data[:, 2]
            else:
                assert False, "Error - %s is an unknown weighting method" % method
        # invert and normalise
        mn, mx = np.nanpercentile(weights, (0, 100))
        weights = mx - weights  # lower values should get higher weight, so flip
    weights /= np.nansum(weights, axis=-1)[..., None]  # sum to 1

    if ascloud:
        weights = hylite.HyCloud(scenes[0].cloud.xyz, bands=np.nan_to_num(weights, nan=0, posinf=0))
        weights.set_band_names([s.name for s in scenes])
    return weights


def blend_scenes(scenes, weights, bands=(0, -1), chunksize=25, trim=False, ooc=True, vb=True):
    """
    Blend together collections of scenes that reference the same underlying cloud to create a fused
    hypercloud.

    Args:
        scenes: a list of scenes to fuse. Data will be extracted from scene.image and pushed to the cloud using
                scene.pmap.
        weights: a numpy array or HyCloud instance containings weighting factors with shape (points,scenes). See
                 get_blend_weights(...) for more details.
        bands: tuple specifying the (min,max) bands of scenes.image to export. Default is all bands (0,-1).
        chunksize: the number of bands to project in any given iteration. Larger numbers are faster but require
                   more RAM.
        trim: True if points with no data mapped to them should be removed.
        ooc: True if hypercloud bands should be created out-of-core and then assembled. This is slower but less RAM
             intensive for large hyperclouds. If this is True, scene.free() will also be called to unload hyperspectral images
             after they have been processed; so please save HyScenes before using!
        vb: True if progress bars should be created.

    Returns:
        a HyCloud instance contain
    """

    # check / format input
    if isinstance(weights, hylite.HyCloud):
        weights = weights.data
    weights /= np.sum(weights, axis=-1)[:, None]  # normalise

    assert len(scenes) == weights.shape[1], "Error: %d scenes != %d weights" % (len(scenes), weights.shape[1])
    assert scenes[0].cloud.point_count() == weights.shape[0], "Error: scene has %d points but weights have %d" % (
    scenes[0].cloud.point_count(), weights.shape[0])

    # make a temp directory
    tmp = mkdtemp()
    err = None

    try:  # wrap everything here in a try, so we can delete the temp folder if issues arise
        # gather data needed for projection
        bands = [scenes[0].image.get_band_index(b) for b in bands]  # convert bands to band indices
        if len(bands) == 2:
            bands = range(bands[0], bands[1] + 1)  # bands is a range of bands
        bands = sorted(set(bands))  # remove duplicates and sort
        wav = [scenes[0].image.get_wavelengths()[b] for b in bands]

        nbands = len(bands)
        mask = np.full(scenes[0].cloud.point_count(), False)  # these become True as valid data is added
        out = scenes[0].cloud.copy(data=False)  # create output cloud

        # loop through scenes and project bands
        for s in scenes:

            assert s.cloud.point_count() == mask.shape[0], "Error: scene %s has %d points, not %d" % (scene.name,
                                                                                                      scene.cloud.point_count(),
                                                                                                      mask.shape[0])
            # make output directory
            pth = os.path.join(tmp, s.name)
            os.makedirs(pth, exist_ok=True)

            # remove zeros
            s.image.set_as_nan(0)

            # push bands to disk
            loop = bands
            if vb:
                loop = tqdm(loop, desc='Projecting scene %s' % s.name, leave=False)
            chunk = []
            for i, b in enumerate(loop):
                chunk.append(b)  # append band to projection queue
                if (len(chunk) == chunksize) or (i == (len(loop) - 1)):
                    data = s.push_to_cloud(chunk).data  # project bands
                    for n in range(data.shape[-1]):  # save them as individual numpy files
                        mask = np.logical_or(mask, np.isfinite(data[:, n]))
                        np.save(os.path.join(pth, 'b%d.npy' % chunk[n]), data[:, n])
                    chunk = []
            if ooc:
                s.free()

        # create output array (n.b. this can be a biiiig malloc)
        out.data = np.zeros((out.point_count(), nbands), dtype=np.float32)

        # combine data
        loop = bands
        if vb:
            loop = tqdm(loop, desc='Blending bands', leave=False)
        for n, b in enumerate(loop):
            sm = np.zeros(mask.shape[0])
            for i, s in enumerate(scenes):
                pth = os.path.join(tmp, s.name)
                out.data[:, n] += np.nan_to_num(np.load(os.path.join(pth, 'b%d.npy' % b)) * weights[:, i])
    except KeyboardInterrupt as inst:
        print("Operation cancelled: cleaning up after KeyboardInterrupt.")
        err = inst
    except Exception as inst:
        print("An unexpected error occured: cleaning up before termination.")
        err = inst  # keep the error to throw later

    # delete temp directory
    shutil.rmtree(tmp)

    # if there was a problem, raise it now!
    if err is not None:
        raise err

    # cleanup point cloud (trim points with no data) and add wavelengths
    if trim:
        out.filter_points(0, np.logical_not(mask))
    out.set_wavelengths(wav)

    return out  # done!
